# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
import weakref
import ast as ast_module
import types
import logging
import pprint

import llvm.core

from numba import pipeline, naming, error, reporting, PY3
from numba.control_flow.control_flow import ControlFlow
from numba.utils import TypedProperty, WriteOnceTypedProperty, NumbaContext
from numba.minivect.minitypes import FunctionType
from numba import functions, symtab
from numba.utility.cbuilder import library
from numba.nodes import metadata
from numba.codegen import translate
from numba.codegen import globalconstants

from numba.intrinsic import default_intrinsic_library
from numba.external import default_external_library
from numba.external.utility import default_utility_library

# ______________________________________________________________________
# Module data

logger = logging.getLogger(__name__)

if PY3:
    NoneType = type(None)
    name_types = str
else:
    NoneType = types.NoneType
    name_types = (str, unicode)

default_pipeline_order = [
    'ast3to2',
    'resolve_templates',
    'validate_signature',
    'update_signature',
    'create_lfunc1',
    'NormalizeASTStage',
    'ControlFlowAnalysis',
    #'ConstFolding',
    'TypeInfer',
    'update_signature',
    'create_lfunc2',
    'TypeSet',
    'dump_cfg',
    'ClosureTypeInference',
    'create_lfunc3',
    'TransformFor',
    'Specialize',
    'RewriteArrayExpressions',
    'SpecializeComparisons',
    'SpecializeSSA',
    'SpecializeClosures',
    'Optimize',
    'Preloader',
    'SpecializeLoops',
    'FixASTLocations',
    'LateSpecializer',
    'ExtensionTypeLowerer',
    'SpecializeFunccalls',
    'SpecializeExceptions',
    'FixASTLocations',
    'cleanup_symtab',
    'CodeGen',
    'LinkingStage',
    'WrapperStage',
    'ErrorReporting',
]

default_type_infer_pipeline_order = [
    'ast3to2',
    'ControlFlowAnalysis',
    'TypeInfer',
]

compile_idx = default_pipeline_order.index('TypeInfer') + 1
default_compile_pipeline_order = default_pipeline_order[compile_idx:]

default_dummy_type_infer_pipeline_order = [
    'ast3to2',
    'TypeInfer',
    'TypeSet',
]

default_numba_lower_pipeline_order = [
    'ast3to2',
    'LateSpecializer',
    'SpecializeFunccalls',
    'SpecializeExceptions',
]


default_numba_wrapper_pipeline_order = default_numba_lower_pipeline_order

default_numba_late_translate_pipeline_order = \
    default_numba_lower_pipeline_order + [
    'CodeGen',
]

# ______________________________________________________________________
# Convenience functions

def insert_stage(pipeline_order, stage, after=None, before=None):
    if after is not None:
        idx = pipeline_order.index(after) + 1
    else:
        idx = pipeline_order.index(before)

    pipeline_order.insert(idx, stage)

# ______________________________________________________________________
# Class definitions

class _AbstractNumbaEnvironment(object):
    '''Used to break circular type dependency between the translation
    and function environments and the top-level NumbaEnvironment.'''

# ______________________________________________________________________

class FunctionErrorEnvironment(object):
    """
    Environment for errors or warnings that occurr during translation of
    a function.
    """

    func = WriteOnceTypedProperty(
        (NoneType, types.FunctionType),
        'Function (or similar) being translated.')

    ast = TypedProperty(
        ast_module.AST,
        'Original Abstract Syntax Tree for the function being translated.')

    source = TypedProperty(
        list, #(str, unicode),
        "Function source code")

    enable_post_mortem = TypedProperty(
        bool,
        "Enable post-mortem debugging for the Numba compiler",
        False|1
    )

    collection = TypedProperty(
        reporting.MessageCollection,
        "Collection of error and warning messages")

    warning_styles = {
        'simple' : reporting.MessageCollection,
        'fancy': reporting.FancyMessageCollection,
    }

    def __init__(self, func, ast, warnstyle):
        self.func = func
        self.ast = ast # copy.deepcopy(ast)

        # Retrieve the source code now
        source_descr = reporting.SourceDescr(func, ast)
        self.source = source_descr.get_lines()

        collection_cls = self.warning_styles[warnstyle]
        self.collection = collection_cls(self.ast, self.source)

    def merge_in(self, parent_error_env):
        """
        Merge error messages into another error environment.
        Useful to propagate error messages for inner functions outwards.
        """
        parent_error_env.collection.messages.extend(self.collection.messages)
        del self.collection.messages[:]

# ______________________________________________________________________

class FunctionEnvironment(object):
    '''State for a function under translation.'''
    # ____________________________________________________________
    # Properties

    numba = WriteOnceTypedProperty(
        _AbstractNumbaEnvironment,
        'Grandparent environment (top-level Numba environment).')

    func = WriteOnceTypedProperty(
        object, 'Function (or similar) being translated.')

    ast = TypedProperty(
        ast_module.AST,
        'Abstract syntax tree for the function being translated.')

    func_signature = TypedProperty(
        FunctionType,
        'Type signature for the function being translated.')

    is_partial = TypedProperty(
        bool,
        "Whether this environment is a partially constructed environment",
        False)

    func_name = TypedProperty(str, 'Target function name.')

    mangled_name = TypedProperty(str, 'Mangled name of compiled function.')

    qualified_name = TypedProperty(str, "Target qualified function name "
                                        "('mymodule.myfunc')")

    llvm_module = TypedProperty(
        llvm.core.Module,
        'LLVM module for this function.  This module is first optimized and '
        'then linked into a global module.  The Python wrapper function goes '
        'directly into the main fat module.')

    error_env = TypedProperty(
        FunctionErrorEnvironment,
        "Error environment for this function.")

    lfunc = TypedProperty(
        (NoneType, llvm.core.Function),
        "Compiled, native, Numba function",
        None)

    lfunc_pointer = TypedProperty(
        (int, long) if not PY3 else int,
        "Pointer to underlying compiled function. Can be used as a callback.",
    )

    link = TypedProperty(
        bool,
        'Flag indicating whether the LLVM function needs to be linked into '
        'the global fast module from LLVMContextManager',
        True)

    wrap = TypedProperty(
        bool,
        'Flag indicating whether the function needs a wrapper function to be '
        'callable from Python.',
        True)

    llvm_wrapper_func = TypedProperty(
        (llvm.core.Function, NoneType),
        'The LLVM wrapper function for the target function.  This is a '
        'wrapper function that accept python object arguments and returns an '
        'object.')

    numba_wrapper_func = TypedProperty(
        object,
        'The Numba wrapper function (see numbafunction.c) for the target '
        'function.  This is a wrapper function that accept python object '
        'arguments and returns an object.')

    symtab = TypedProperty(
        (symtab.Symtab, dict),
        'A map from local variable names to symbol table variables for all '
        'local variables. '
        '({ "local_var_name" : numba.symtab.Variable(local_var_type) })')

    function_globals = TypedProperty(
        (dict, NoneType),
        "Globals dict of the function",)

    locals = TypedProperty(
        dict,
        'A map from local variable names to types.  Used to handle the locals '
        'keyword argument to the autojit decorator. '
        '({ "local_var_name" : local_var_type } for @autojit(locals=...))')

    template_signature = TypedProperty(
        object, # FIXME
        'Template signature for @autojit.  E.g. T(T[:, :]).  See '
        'numba.typesystem.templatetypes.')

    ast_metadata = TypedProperty(
        object,
        'Metadata for AST nodes of the function being compiled.')

    flow = TypedProperty(
        (NoneType, ControlFlow),
        "Control flow graph. See numba.control_flow.",
        default=None)

    # FIXME: Get rid of this.  See comment for translator property,
    # below.
    cfg_transform = TypedProperty(
        object, # Should be ControlFlowAnalysis.
        'The Control Flow Analysis transform object '
        '(control_flow.ControlFlowAnalysis). Set during the cfg pass.')

    # FIXME: Get rid of this property; pipeline stages are users and
    # transformers of the environment.  Any state needed beyond a
    # given stage should be put in the environment instead of keeping
    # around the whole transformer.
    # TODO: Create linking stage
    translator = TypedProperty(
        object, # FIXME: Should be LLVMCodeGenerator, but that causes
                # module coupling.
        'The code generator instance used to generate the target LLVM '
        'function.  Set during the code generation pass, and used for '
        'after-the-fact wrapper generation.')

    is_closure = TypedProperty(
        bool,
        'Flag indicating if the current function under translation is a '
        'closure or not.',
        False)

    closures = TypedProperty(
        dict, 'Map from ast nodes to closures.')

    closure_scope = TypedProperty(
        (dict, NoneType),
        'Collective symtol table containing all entries from outer '
        'functions.')

    need_closure_wrapper = TypedProperty(
        bool, "Whether this closure needs a Python wrapper function",
        default=True,
    )

    refcount_args = TypedProperty(
        bool, "Whether to use refcounting for the function arguments", True)

    warn = TypedProperty(
        bool,
        'Flag that enables control flow warnings on a per-function level.',
        True)

    warnstyle = TypedProperty(
        str if PY3 else basestring,
        'Warning style, currently available: simple, fancy',
        default='fancy'
    )

    kwargs = TypedProperty(
        dict,
        'Additional keyword arguments.  Deprecated, but kept for backward '
        'compatibility.')

    # ____________________________________________________________
    # Methods

    def __init__(self, *args, **kws):
        self.init(*args, **kws)

    def init(self, parent, func, ast, func_signature,
             name=None, qualified_name=None,
             mangled_name=None,
             llvm_module=None, wrap=True, link=True,
             symtab=None,
             error_env=None, function_globals=None, locals=None,
             template_signature=None, cfg_transform=None,
             is_closure=False, closures=None, closure_scope=None,
             refcount_args=True,
             ast_metadata=None, warn=True, warnstyle='fancy',
             **kws):

        self.parent = parent
        self.numba = parent.numba
        self.func = func
        self.ast = ast
        self.func_signature = func_signature

        if name is None:
            if self.func:
                name = self.func.__name__
            else:
                name = self.ast.name

        if self.func and self.func.__module__:
            qname = '.'.join([self.func.__module__, name])
        else:
            qname = name

        if mangled_name is None:
            mangled_name = naming.specialized_mangle(qname,
                                                     self.func_signature.args)

        self.func_name = name
        self.mangled_name = mangled_name
        self.qualified_name = qualified_name or name
        self.llvm_module = (llvm_module if llvm_module
                                 else self.numba.llvm_context.module)
        self.wrap = wrap
        self.link = link
        self.llvm_wrapper_func = None
        self.symtab = symtab if symtab is not None else {}

        self.error_env = error_env or FunctionErrorEnvironment(self.func,
                                                               self.ast,
                                                               warnstyle)

        if function_globals is not None:
            self.function_globals = function_globals
        else:
            self.function_globals = self.func.__globals__

        self.locals = locals if locals is not None else {}
        self.template_signature = template_signature
        self.cfg_transform = cfg_transform
        self.is_closure = is_closure
        self.closures = closures if closures is not None else {}
        self.closure_scope = closure_scope

        self.refcount_args = refcount_args

        if ast_metadata is not None:
            self.ast_metadata = ast_metadata
        else:
            self.ast_metadata = metadata.create_metadata_env()

        self.warn = warn
        self.warnstyle = warnstyle
        self.kwargs = kws

    def getstate(self):
        state = dict(
            parent=self.parent,
            func=self.func,
            ast=self.ast,
            func_signature=self.func_signature,
            name=self.func_name,
            mangled_name=self.mangled_name,
            qualified_name=self.qualified_name,
            llvm_module=self.llvm_module,
            wrap=self.wrap,
            link=self.link,
            symtab=self.symtab,
            function_globals=self.function_globals,
            locals=self.locals,
            template_signature=self.template_signature,
            cfg_transform=self.cfg_transform,
            is_closure=self.is_closure,
            closures=self.closures,
            closure_scope=self.closure_scope,
            warn=self.warn,
            warnstyle=self.warnstyle,
        )
        return state

    def inherit(self, **kwds):
        """
        Inherit from a parent FunctionEnvironment (e.g. to run pipeline stages
        on a subset of the AST).
        """
        # TODO: link these things together
        state = self.getstate()
        state.update(kwds)
        return type(self)(**state)

    @property
    def func_doc(self):
        if self.func is not None:
            return self.func.__doc__
        else:
            return ast_module.get_docstring(self.ast)

# ______________________________________________________________________

class TranslationEnvironment(object):
    '''State for a given translation.'''
    # ____________________________________________________________
    # Properties

    numba = TypedProperty(_AbstractNumbaEnvironment, 'Parent environment')

    crnt = TypedProperty(
        (FunctionEnvironment, NoneType),
        'The environment corresponding to the current function under '
        'translation.')

    stack = TypedProperty(
        list,
        'A stack consisting of FunctionEnvironment instances.  Used to '
        'manage lexical closures.')

    functions = TypedProperty(
        dict,
        'A map from target function names that are under compilation to their '
        'corresponding FunctionEnvironments')

    func_envs = TypedProperty(
        weakref.WeakKeyDictionary,
        "Map from root AST nodes to FunctionEnvironment objects."
        "This allows tracking environments of partially processed ASTs.")

    nopython = TypedProperty(
        bool,
        'Flag used to indicate if calls to the Python C API are permitted or '
        'not during code generation.',
        False)

    allow_rebind_args = TypedProperty(
        bool,
        'Flag indicating whether the type of arguments may be overridden for '
        '@jit functions.  This is always true (except for in tests perhaps!)',
        True)

    warn = TypedProperty(
        bool,
        'Flag that enables control flow warnings. FunctionEnvironment inherits '
        'this unless overridden.',
        True)

    is_pycc = TypedProperty(
        bool,
        'Flag that tells us whether this function is being exported with pycc.',
        False)

    # ____________________________________________________________
    # Methods

    def __init__(self, parent, **kws):
        self.numba = parent
        self.crnt = None
        self.stack = [(kws, None)]
        self.functions = {}
        self.func_envs = weakref.WeakKeyDictionary()
        self.set_flags(**kws)

    def set_flags(self, **kws):
        self.nopython = kws.get('nopython', False)
        self.allow_rebind_args = kws.get('allow_rebind_args', True)
        self.warn = kws.get('warn', True)
        self.is_pycc = kws.get('is_pycc', False)

    def get_or_make_env(self, func, ast, func_signature, **kwds):
        if ast not in self.func_envs:
            kwds.setdefault('warn', self.warn)
            func_env = self.numba.FunctionEnvironment(
                    self, func, ast, func_signature, **kwds)
            self.func_envs[ast] = func_env
        else:
            func_env = self.func_envs[ast]
            if func_env.is_partial:
                state = func_env.partial_state
            else:
                state = func_env.getstate()

            state.update(kwds, func=func, ast=ast,
                               func_signature=func_signature)
            func_env.init(self, **state)

        return func_env

    def get_env(self, ast):
        if ast in self.func_envs:
            return self.func_envs[ast]
        else:
            return None

    def make_partial_env(self, ast, **kwds):
        """
        Create a partial environment for a function that only initializes
        the given attributes.

        Later attributes will override existing attributes.
        """
        if ast in self.func_envs:
            func_env = self.func_envs[ast]
        else:
            func_env = self.numba.FunctionEnvironment.__new__(
                    self.numba.FunctionEnvironment)
            func_env.is_partial = True
            func_env.partial_state = kwds

            for key, value in kwds.iteritems():
                setattr(func_env, key, value)

            self.func_envs[ast] = func_env
            func_env.ast = ast

        return func_env

    def push(self, func, ast, func_signature, **kws):
        func_env = self.get_or_make_env(func, ast, func_signature, **kws)
        return self.push_env(func_env, **kws)

    def push_env(self, func_env, **kws):
        self.set_flags(**kws)
        self.crnt = func_env
        self.stack.append((kws, self.crnt))
        self.functions[self.crnt.func_name] = self.crnt
        self.func_envs[func_env.ast] = func_env
        if self.numba.debug:
            logger.debug('stack=%s\ncrnt=%r (%r)', pprint.pformat(self.stack),
                         self.crnt, self.crnt.func if self.crnt else None)
        return self.crnt

    def pop(self):
        ret_val = self.stack.pop()
        kws, self.crnt = self.stack[-1]
        self.set_flags(**kws)
        if self.numba.debug:
            logger.debug('stack=%s\ncrnt=%r (%r)', pprint.pformat(self.stack),
                         self.crnt, self.crnt.func if self.crnt else None)
        return ret_val

# ______________________________________________________________________

class TranslationContext(object):
    """Context manager for handling a translation.  Pushes a
    FunctionEnvironment input onto the given translation environment's
    stack, and pops it when leaving the translation context.
    """
    def __init__(self, env, *args, **kws):
        self.translation_environment = env.translation
        self.args = args
        self.kws = kws

    def __enter__(self):
        return self.translation_environment.push(*self.args, **self.kws)

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.translation_environment.pop()

# ______________________________________________________________________

class PyccEnvironment(object):
    '''pycc environment

    Includes flags, and modules for exported functions.
    '''
    # ____________________________________________________________
    # Properties

    wrap_exports = TypedProperty(
        bool,
        'Boolean flag used to indicate that Python wrappers should be '
        'generated for exported functions.')

    function_signature_map = TypedProperty(
        dict,
        'Map from function names to type signatures for the translated '
        'function (used for header generation).')

    function_module_map = TypedProperty(
        dict,
        'Map from function names to LLVM modules that define the translated '
        'function.')

    function_wrapper_map = TypedProperty(
        dict,
        'Map from function names to tuples containing LLVM wrapper functions '
        'and LLVM modules that define the wrapper function.')

    # ____________________________________________________________
    # Methods

    def __init__(self, wrap_exports=False, *args, **kws):
        self.reset(wrap_exports, *args, **kws)

    def reset(self, wrap_exports=False, *args, **kws):
        '''Clear the current set of exported functions.'''
        self.wrap_exports = wrap_exports
        self.function_signature_map = {}
        self.function_module_map = {}
        self.function_wrapper_map = {}

# ______________________________________________________________________

class NumbaEnvironment(_AbstractNumbaEnvironment):
    '''Defines global state for a Numba translator. '''
    # ____________________________________________________________
    # Properties

    name = TypedProperty(name_types, "Name of the environment.")

    pipelines = TypedProperty(
        dict, 'Map from entry point names to PipelineStages.')

    pipeline_stages = TypedProperty(
        types.ModuleType,
        'Namespace for pipeline stages.  Initially set to the numba.pipeline '
        'module.',
        pipeline)

    default_pipeline = TypedProperty(
        str,
        'Default entry point name.  Used to index into the "pipelines" map.',
        default='numba')

    context = TypedProperty(
        NumbaContext,
        'Defines a global typing context for handling promotions and type '
        'representations.')

    specializations = TypedProperty(
        functions.FunctionCache, 'Cache for previously specialized functions.')

    exports = TypedProperty(
        PyccEnvironment, 'Translation environment for pycc usage')

    debug = TypedProperty(
        bool,
        'Global flag indicating verbose debugging output should be enabled.',
        False)

    debug_coercions = TypedProperty(
        bool,
        'Flag for checking type coercions done during late specialization.',
        False)

    stage_checks = TypedProperty(
        bool,
        'Global flag for enabling detailed checks in translation pipeline '
        'stages.',
        False)

    translation = TypedProperty(
        TranslationEnvironment,
        'Current translation environment, specific to the current pipeline '
        'being run.')

    llvm_context = TypedProperty(
        translate.LLVMContextManager,
        "Manages the global LLVM module and linkages of new translations."
    )

    constants_manager = TypedProperty(
        globalconstants.LLVMConstantsManager,
        "Holds constant values in an LLVM module.",
        default=globalconstants.LLVMConstantsManager(),
    )

    # ____________________________________________________________
    # Class members

    environment_map = {}

    TranslationContext = TranslationContext
    TranslationEnvironment = TranslationEnvironment
    FunctionEnvironment = FunctionEnvironment

    # ____________________________________________________________
    # Methods

    def __init__(self, name, *args, **kws):
        self.name = name
        actual_default_pipeline = pipeline.ComposedPipelineStage(
            default_pipeline_order)
        self.pipelines = {
            self.default_pipeline : actual_default_pipeline,
            'type_infer' : pipeline.ComposedPipelineStage(
                default_type_infer_pipeline_order),
            'dummy_type_infer' : pipeline.ComposedPipelineStage(
                default_dummy_type_infer_pipeline_order),
            'compile' : pipeline.ComposedPipelineStage(
                default_compile_pipeline_order),
            'wrap_func' : pipeline.ComposedPipelineStage(
                default_numba_wrapper_pipeline_order),
            'lower' : pipeline.ComposedPipelineStage(
                default_numba_lower_pipeline_order),
            'late_translate' : pipeline.ComposedPipelineStage(
                default_numba_late_translate_pipeline_order),
            }
        self.context = NumbaContext()
        self.specializations = functions.FunctionCache(env=self)
        self.exports = PyccEnvironment()
        self.translation = self.TranslationEnvironment(self)
        self.debug = logger.getEffectiveLevel() < logging.DEBUG

        # FIXME: NumbaContext has up to now been used as a stand in
        # for NumbaEnvironment, so the following member definitions
        # should be moved into the environment, and the code that uses
        # them should be updated.
        context = self.context
        context.env = self
        context.numba_pipeline = actual_default_pipeline
        context.function_cache = self.specializations
        context.intrinsic_library = default_intrinsic_library(context)
        context.external_library = default_external_library(context)
        context.utility_library = default_utility_library(context)
        self.llvm_context = translate.LLVMContextManager()

    def link_cbuilder_utilities(self):
        self.context.cbuilder_library = library.CBuilderLibrary()
        self.context.cbuilder_library.declare_registered(self)

        # Link modules
        self.context.cbuilder_library.link(self.llvm_context.module)

    @classmethod
    def get_environment(cls, environment_key = None, *args, **kws):
        '''
        Given an optional key, return the global Numba environment for
        that key.  If no key is given, return the default global
        environment.

        Note that internally, the default environment is mapped to None.
        '''
        if environment_key in cls.environment_map:
            ret_val = cls.environment_map[environment_key]
        else:
            ret_val = cls(environment_key or 'numba', *args, **kws)
            cls.environment_map[environment_key] = ret_val
        return ret_val

    @property
    def crnt(self):
        return self.translation.crnt

    def get_pipeline(self, pipeline_name=None):
        '''Convenience function for getting a pipeline object (which
        should be a callable object that accepts (ast, env) arguments,
        and returns an ast).'''
        if pipeline_name is None:
            pipeline_name = self.default_pipeline
        return self.pipelines[pipeline_name]

    def get_or_add_pipeline(self, pipeline_name=None, pipeline_ctor=None):
        if pipeline_name is None:
            pipeline_name = self.default_pipeline
        if pipeline_name in self.pipelines:
            pipeline_obj = self.pipelines[pipeline_name]
        else:
            pipeline_obj = self.pipelines[pipeline_name] = pipeline_ctor()
        return pipeline_obj

    def __repr__(self):
        return "NumbaEnvironment(%s)" % self.name

# ______________________________________________________________________
# Main (self-test) routine

def main(*args):
    import numba as nb
    test_ast = ast_module.parse('def test_fn(a, b):\n  return a + b\n\n',
                                   '<string>', 'exec')
    exec(compile(test_ast, '<string>', 'exec'))
    test_fn_ast = test_ast.body[-1]
    test_fn_sig = nb.double(nb.double, nb.double)
    test_fn_sig.name = test_fn.__name__
    env = NumbaEnvironment.get_environment()
    with TranslationContext(env, test_fn, test_fn_ast, test_fn_sig):
        env.get_pipeline()(test_fn_ast, env)
    assert env.pipeline_stages == pipeline

if __name__ == "__main__":
    import sys
    main(*sys.argv[1:])



# -*- coding: utf-8 -*-
"""
Struct and complex nodes.

Structs are allocated on the stack, and not mutated as values. This is
because mutations are attribute or index assignments, which are not
recognized as variable assignments. Hence mutation cannot propagate new
values. So we mutate what we have on the stack.
"""
from __future__ import print_function, division, absolute_import

from numba.nodes import *

def struct_type(type):
    if type.is_reference:
        type = type.referenced_type

    return type

class StructAttribute(ExprNode):
    # expr : = StructAttribute(expr, string, expr_context, Type, metadata)
    # metadata := StructAttribute | ComplexAttribute
    _fields = ['value']

    def __init__(self, value, attr, ctx, type, **kwargs):
        super(StructAttribute, self).__init__(**kwargs)
        self.value = value
        self.attr = attr
        self.ctx = ctx
        self.struct_type = type

        type = struct_type(type)
        self.attr_type = type.fielddict[attr]

        self.type = self.attr_type
        self.variable = Variable(self.type, promotable_type=False)

    @property
    def field_idx(self):
        fields = struct_type(self.struct_type).fields
        return fields.index((self.attr, self.attr_type))

class StructVariable(ExprNode):
    """
    Tells the type inferencer that the node is actually a valid struct that
    we can mutate. For instance

        func().a = 2

    is wrong if func() returns a struct by value. So we only allow references
    like struct.a = 2 and array[i].a = 2.
    """

    _fields = ['node']

    def __init__(self, node, **kwargs):
        super(StructVariable, self).__init__(**kwargs)
        self.node = node
        self.type = node.type

class ComplexNode(ExprNode):
    _fields = ['real', 'imag']

    type = complex128
    variable = Variable(type)

    def __init__(self, real, imag):
        self.real = real
        self.imag = imag

class ComplexAttributeNode(ExprNode):

    _fields = ["value"]

    def __init__(self, value, attr):
        self.value = value
        self.attr = attr
        self.type = value.type.base_type
        self.variable = Variable(self.type)



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

from numba.exttypes.entrypoints import  (jit_extension_class,
                                         autojit_extension_class,
                                         autojit_class_wrapper)

__all__ = ['autojit', 'jit', 'export', 'exportmany']

import types
import logging
import inspect

from numba import *
from numba import typesystem, numbawrapper
from numba import  functions
from .minivect import minitypes
from numba.utils import  process_signature
from numba.codegen import llvmwrapper
from numba import environment
import llvm.core as _lc
from numba.wrapping import compiler

logger = logging.getLogger(__name__)

environment.NumbaEnvironment.get_environment().link_cbuilder_utilities()

if PY3:
    CLASS_TYPES = type
else:
    CLASS_TYPES = (type, types.ClassType)

#------------------------------------------------------------------------
# PyCC decorators
#------------------------------------------------------------------------

def _internal_export(env, function_signature, backend='ast', **kws):
    def _iexport(func):
        if backend == 'bytecode':
            raise NotImplementedError(
               'Bytecode translation has been removed for exported functions.')
        else:
            name = function_signature.name
            llvm_module = _lc.Module.new('export_%s' % name)
            if not hasattr(func, 'live_objects'):
                func.live_objects = []
            func._is_numba_func = True
            func_ast = functions._get_ast(func)
            # FIXME: Hacked "mangled_name" into the translation
            # environment.  Should do something else.  See comment in
            # codegen.translate.LLVMCodeGenerator.__init__().
            with environment.TranslationContext(
                    env, func, func_ast, function_signature,
                    name=name, llvm_module=llvm_module,
                    mangled_name=name,
                    link=False, wrap=False,
                    is_pycc=True) as func_env:
                pipeline = env.get_pipeline()
                func_ast.pipeline = pipeline
                pipeline(func_ast, env)
                exports_env = env.exports
                exports_env.function_signature_map[name] = function_signature
                exports_env.function_module_map[name] = llvm_module
                if not exports_env.wrap_exports:
                    exports_env.function_wrapper_map[name] = None
                else:
                    wrapper_tup = llvmwrapper.build_wrapper_module(env)
                    exports_env.function_wrapper_map[name] = wrapper_tup
        return func
    return _iexport

def export(signature, env_name=None, env=None, **kws):
    """
    Construct a decorator that takes a function and exports one

    A signature is a string with

    name ret_type(arg_type, argtype, ...)
    """
    if env is None:
        env = environment.NumbaEnvironment.get_environment(env_name)
    return _internal_export(env, process_signature(signature), **kws)

def exportmany(signatures, env_name=None, env=None, **kws):
    """
    A Decorator that exports many signatures for a single function
    """
    if env is None:
        env = environment.NumbaEnvironment.get_environment(env_name)
    def _export(func):
        for signature in signatures:
            tocall = _internal_export(env, process_signature(signature), **kws)
            tocall(func)
        return func
    return _export

#------------------------------------------------------------------------
# Compilation Entry Points
#------------------------------------------------------------------------

# TODO: Redo this entire module

def compile_function(env, func, argtypes, restype=None, **kwds):
    """
    Compile a python function given the argument types. Compile only
    if not compiled already, and only if it is registered to the function
    cache.

    Returns a triplet of (signature, llvm_func, python_callable)
    `python_callable` is the wrapper function (NumbaFunction).
    """
    function_cache = env.specializations

    # For NumbaFunction, we get the original python function.
    func = getattr(func, 'py_func', func)

    # get the compile flags
    flags = None # stub

    # Search in cache
    result = function_cache.get_function(func, argtypes, flags)
    if result is not None:
        sig, lfunc, pycall = result
        return sig, lfunc, pycall

    # Compile the function
    from numba import pipeline

    compile_only = getattr(func, '_numba_compile_only', False)
    kwds['compile_only'] = kwds.get('compile_only', compile_only)

    assert kwds.get('llvm_module') is None, kwds.get('llvm_module')

    func_env = pipeline.compile2(env, func, restype, argtypes, **kwds)

    function_cache.register_specialization(func_env)
    return (func_env.func_signature,
            func_env.lfunc,
            func_env.numba_wrapper_func)


def _autojit(template_signature, target, nopython, env_name=None, env=None,
             **flags):
    if env is None:
        env = environment.NumbaEnvironment.get_environment(env_name)

    def _autojit_decorator(f):
        """
        Defines a numba function, that, when called, specializes on the input
        types. Uses the AST translator backend. For the bytecode translator,
        use @autojit.
        """

        if isinstance(f, CLASS_TYPES):
            compiler_cls = compiler.ClassCompiler
            wrapper = autojit_class_wrapper
        else:
            compiler_cls = compiler.FunctionCompiler
            wrapper = autojit_wrappers[(target, 'ast')]

        env.specializations.register(f)
        cache = env.specializations.get_autojit_cache(f)

        flags['target'] = target
        compilerimpl = compiler_cls(env, f, nopython, flags, template_signature)
        numba_func = wrapper(f, compilerimpl, cache)

        return numba_func

    return _autojit_decorator

def autojit(template_signature=None, backend='ast', target='cpu',
            nopython=False, locals=None, **kwargs):
    """
    Creates a function that dispatches to type-specialized LLVM
    functions based on the input argument types.  If no specialized
    function exists for a set of input argument types, the dispatcher
    creates and caches a new specialized function at call time.
    """
    if template_signature and not isinstance(template_signature, minitypes.Type):
        if callable(template_signature):
            func = template_signature
            return autojit(backend='ast', target=target,
                           nopython=nopython, locals=locals, **kwargs)(func)
        else:
            raise Exception("The autojit decorator should be called: "
                            "@autojit()")

    if backend == 'bytecode':
        return _not_implemented
    else:
        return _autojit(template_signature, target, nopython,
                        locals=locals, **kwargs)

def _jit(restype=None, argtypes=None, nopython=False,
         _llvm_module=None, env_name=None, env=None, **kwargs):
    if env is None:
        env = environment.NumbaEnvironment.get_environment(env_name)
    def _jit_decorator(func):
        if isinstance(func, CLASS_TYPES):
            cls = func
            kwargs.update(env_name=env_name)
            return jit_extension_class(cls, kwargs, env)

        argtys = argtypes
        if func.__code__.co_argcount == 0 and argtys is None:
            argtys = []

        assert argtys is not None
        env.specializations.register(func)

        assert kwargs.get('llvm_module') is None # TODO link to user module
        assert kwargs.get('llvm_ee') is None, "Engine should never be provided"
        sig, lfunc, wrapper = compile_function(env, func, argtys, restype=restype,
                                    nopython=nopython, **kwargs)
        return numbawrapper.create_numba_wrapper(func, wrapper, sig, lfunc)

    return _jit_decorator

def _not_implemented(*args, **kws):
    raise NotImplementedError('Bytecode backend is no longer supported.')

jit_targets = {
    ('cpu', 'bytecode') : _not_implemented,
    ('cpu', 'ast') : _jit,
}

autojit_wrappers = {
    ('cpu', 'bytecode') : _not_implemented,
    ('cpu', 'ast')      : numbawrapper.NumbaSpecializingWrapper,
}

def jit(restype=None, argtypes=None, backend='ast', target='cpu', nopython=False,
        **kws):
    """
    Compile a function given the input and return types.

    There are multiple ways to specify the type signature:

    * Using the restype and argtypes arguments, passing Numba types.

    * By constructing a Numba function type and passing that as the
      first argument to the decorator.  You can create a function type
      by calling an exisiting Numba type, which is the return type,
      and the arguments to that call define the argument types.  For
      example, ``f8(f8)`` would create a Numba function type that
      takes a single double-precision floating point value argument,
      and returns a double-precision floating point value.

    * As above, but using a string instead of a constructed function
      type.  Example: ``jit("f8(f8)")``.

    If backend='bytecode' the bytecode translator is used, if
    backend='ast' the AST translator is used.  By default, the AST
    translator is used.  *Note that the bytecode translator is
    deprecated as of the 0.3 release.*
    """
    kws.update(nopython=nopython, backend=backend)
    if isinstance(restype, CLASS_TYPES):
        cls = restype
        env = kws.pop('env', None) or environment.NumbaEnvironment.get_environment(
                                                         kws.get('env_name', None))
        return jit_extension_class(cls, kws, env)

    # Called with f8(f8) syntax which returns a dictionary of argtypes and restype
    if isinstance(restype, minitypes.FunctionType):
        if argtypes is not None:
            raise TypeError("Cannot use both calling syntax and argtypes keyword")
        argtypes = restype.args
        restype = restype.return_type
    # Called with a string like 'f8(f8)'
    elif isinstance(restype, str) and argtypes is None:
        signature = process_signature(restype, kws.get('name', None))
        name, restype, argtypes = (signature.name, signature.return_type,
                                   signature.args)
        if name is not None:
            kws['func_name'] = name
    if restype is not None:
        kws['restype'] = restype
    if argtypes is not None:
        kws['argtypes'] = argtypes

    return jit_targets[target, backend](**kws)




# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
import sys
import ast
from . import asdl
from collections import defaultdict, namedtuple
import contextlib

def load(name):
    '''Load a ASDL Schema by name; e.g. "Python.asdl".
        
    Returns a Schema object.

    This tries to load from the version-specific directory, first.
    If it failed, it loads from the common-directory.
    '''
    python_asdl = asdl.load(name)
    schblr = SchemaBuilder()
    schblr.visit(python_asdl)
    return schblr.schema

class _rule(namedtuple('rule', ['kind', 'fields'])):
    __slots__ = ()

    SUM = 0
    PROD = 1

    @classmethod
    def sum(cls, fields):
        return cls(kind=cls.SUM, fields=fields)

    @classmethod
    def product(cls, fields):
        return cls(kind=cls.PROD, fields=fields)

    @property
    def is_sum(self):
        return self.kind == self.SUM

    @property
    def is_product(self):
        return self.kind == self.PROD

class _debuginfo_t(namedtuple("_debuginfo_t", ['node', 'field', 'offset'])):
    __slots__ = ()

    def __str__(self):
        '''Create string reprentation to be used in SchemaError
        '''
        if self.field is not None:
            if self.offset is not None:
                return "At %s.%s[%d]" % (self.node, self.field, self.offset)
            else:
                return "At %s.%s" % (self.node, self.field)
        else:
            return "At %s" % (self.node)

def _debuginfo(node, field=None, offset=None):
    assert field or not offset
    return _debuginfo_t(node=node, field=field, offset=offset)

class SchemaError(Exception):
    def __init__(self, ctxt, msg):
        super(SchemaError, self).__init__("%s: %s" % (ctxt, msg))

class Schema(object):
    '''A Schema object that is used to verify against an AST.
    
    It is built from SchemaBuilder
        
    Usage:

        schema.verify(ast)
        schema.verify(ast, context=SchemaContext())
    '''
    def __init__(self, name):
        # name of the asdl module
        self.name = name
        # a dictionary of {type name -> fields}
        self.types = defaultdict(list)
        # a dictionary of {definition name -> _rule}
        self.dfns = {}

    def verify(self, ast, context=None):
        '''Check against an AST raises SchemaError upon error.
        
        ast --- The ast being verified
        context --- [optional] a SchemaContext.
        '''
        context = context if context is not None else SchemaContext()
        return SchemaVerifier(self, context).visit(ast)

    def debug(self):
        print(("Schema %s" % self.name))
        for k, fs in self.types.items():
            print(('    --', k, fs))
        for k, fs in self.dfns.items():
            if fs.is_sum():
                print(('   ', k))
                for f in fs.fields:
                    print(('       |', f))
            else:
                print(('   ', k, '=', ', '.join(map(str, fs.fields))))

class SchemaContext(object):
    '''Keep information about context:
        - builtin type handlers
        
    User may expand the builtin type handlers.  See `builtin_handlers`.
        
    '''
    def __init__(self):
        self.__builtins = {}
        self._add_default_handler()

    def _add_default_handler(self):
        self.builtin_handlers['identifier'] = _verify_identifier
        self.builtin_handlers['int'] = _verify_int
        self.builtin_handlers['string'] = _verify_string
        self.builtin_handlers['object'] = _verify_object
        self.builtin_handlers['bool'] = _verify_bool

    @property
    def builtin_handlers(self):
        '''A dictionary of type name -> handler
            
        A handler is just a callable like the following:
            
            def handler(value):
                return is_value_valid(value) # returns boolean
        '''
        return self.__builtins


class SchemaVerifier(ast.NodeVisitor):
    '''A internal class that implement a the verification logic.
    '''
    def __init__(self, schema, context):
        '''
        schema --- a Schema object that defines a valid AST.
        context --- a SchemaConctext object.
        '''
        self.schema = schema
        self.context = context
        self._debug_context = None
    
    def visit(self, node):
        '''Start verification at the node.
        
        Verification can begin at any AST node.  Can it will recursively
        verify each and every subtree.
        '''
        current = self._get_type(node)
        self._visit(node, current)

    def _visit(self, node, current):
        nodename = type(node).__name__
        # traverse the children

        for field in current:
            with self._new_debug_context(node=nodename, field=field.name):
                value = getattr(node, str(field.name), None)
                if getattr(field, 'seq', False):
                    # is sequence?
                    children = getattr(node, str(field.name), None)
                    if children is None:
                        raise SchemaError(self._debug_context, "Missing field")
                    elif not _is_iterable(children):
                        raise SchemaError(self._debug_context,
                                          "Field must be iterable")

                    for offset, child in enumerate(children):
                        with self._new_debug_context(node=nodename,
                                                     field=field.name,
                                                     offset=offset):
                            self._sentry_dfn(field.type, child)

                elif value is None:
                    if not getattr(field, 'opt', False):
                        raise SchemaError(self._debug_context, "Missing field")
                    else:
                        pass
                elif value is not None:
                    self._sentry_dfn(field.type, value)

                else:
                    assert False, field

    @contextlib.contextmanager
    def _new_debug_context(self, **kws):
        # push
        old = self._debug_context
        self._debug_context = _debuginfo(**kws)
        yield
        # pop
        self._debug_context = old

    def _sentry_child(self, child, subtypes):
        childname = type(child).__name__
        if childname not in subtypes:
            raise SchemaError(self._debug_context, "Cannot be a %s" % childname)

    def _sentry_dfn(self, name, value):
        name = str(name)

        if name in self.context.builtin_handlers:
            # is a builtin type?
            handler = self.context.builtin_handlers[name]
            if not handler(value):
                raise SchemaError(self._debug_context,
                                  "Expected %s but got %s" % \
                                  (name, type(value)))
        else:
            # not a builtin type?
            rule = self._get_dfn(name)
            if rule.is_sum:
                self._sentry_child(value, rule.fields)
                self.visit(value)
            else:
                assert rule.is_product
                name0 = type(value).__name__
                if name0 != name:
                    raise SchemaError(self._debug_context,
                                      "Expecting %s but got %s" % \
                                      (name, name0))

                self._visit(value, rule.fields)

    def _get_dfn(self, name):
        ret = self.schema.dfns.get(str(name))
        if ret is None:
            raise SchemaError(self._debug_context,
                              "Missing definition for '%s' in the ASDL" % name)
        return ret

    def _get_type(self, cls_or_name):
        name = (cls_or_name
                if isinstance(cls_or_name, str)
                else type(cls_or_name).__name__)
        ret = self.schema.types.get(name)
        if ret is None:
            raise SchemaError(self._debug_context,
                              "Unknown AST node type: %s" % name)
        return ret

class SchemaBuilder(asdl.VisitorBase):
    '''A single instance of SchemaBuilder can be used build different 
    Schema from different ASDL.

    Usage:
        schblr = SchemaBuilder()
        schblr.visit(some_asdl)
        schema = schblr.schema      # get the schema object
        
    NOTE:
        - It ignore the attributes of the nodes.
    '''
    def __init__(self):
        super(SchemaBuilder, self).__init__()

    def visitModule(self, mod):
        self.__schema = Schema(str(mod.name))
        for dfn in mod.dfns:
            self.visit(dfn)

    def visitType(self, type):
        self.visit(type.value, str(type.name))

    def visitSum(self, sum, name):
        self.schema.dfns[str(name)] = _rule.sum([str(t.name)
                                                 for t in sum.types])
        for t in sum.types:
            self.visit(t, name, sum.attributes)

    def visitConstructor(self, cons, name, attr=[]):
        typename = str(cons.name)
        fields = self.schema.types[typename]
        for f in cons.fields:
            fields.append(f)
        for f in attr:
            fields.append(f)


    def visitField(self, field, name):
        key = str(field.type)

    def visitProduct(self, prod, name):
        assert not hasattr(prod, 'attributes')
        self.schema.dfns[str(name)] = _rule.product(prod.fields)
        for f in prod.fields:
            self.visit(f, name)

    @property
    def schema(self):
        return self.__schema

#
# Builtin types handler
#

def _verify_identifier(value):
    return isinstance(value, str)

def _verify_int(value):
    return isinstance(value, int) or isinstance(value, long)

def _verify_string(value):
    return isinstance(value, str)

def _verify_object(value):
    return isinstance(value, object)

def _verify_bool(value):
    return isinstance(value, bool)

#
#  Utilities
#
def _is_iterable(value):
    try:
        iter(value)
    except TypeError:
        return False
    else:
        return True



import sys
from numba import *

@jit
class Base(object):

    @void(double)
    def __init__(self, myfloat):
        self.value = myfloat

    @double()
    def getvalue(self):
        "Return value"
        return self.value

@jit
class Derived1(Base):

    @void(double)
    def __init__(self, value):
        self.value = value
        self.value2 = double(2.0)

def test_sizeof_extra_attr():
    base = Base(10.0)
    derived = Derived1(10.0)
    base_size = sys.getsizeof(base)
    derived_size = sys.getsizeof(derived)
    assert base_size + 8 == derived_size, (base_size, derived_size)

@jit
class Derived2(Base):

    @double()
    def getvalue(self):
        return self.value

def test_sizeof_extra_method():
    base_size = sys.getsizeof(Base(10.0))
    derived_size = sys.getsizeof(Derived2(10.0))
    assert base_size == derived_size, (base_size, derived_size)


if __name__ == '__main__':
    test_sizeof_extra_attr()
    test_sizeof_extra_method()



# -*- coding: utf-8 -*-
"""
Error reporting. Used by the CFA and by each FunctionEnvironment,
which can collect errors and warnings and issue them after failed or
successful compilation.
"""
from __future__ import print_function, division, absolute_import

import sys
import inspect

from numba import error

def getpos(node):
    try:
        return node.lineno, node.col_offset
    except:
        return 0, 0
# ______________________________________________________________________

class SourceDescr(object):
    """
    Source code descriptor.
    """

    def __init__(self, func, ast):
        self.func = func
        self.ast = ast

    def get_lines(self):
        source = None
        if self.func:
            try:
                source = inspect.getsource(self.func)
            except EnvironmentError:
                pass

        if source is None:
            try:
                from meta import asttools
                source = asttools.dump_python_source(self.ast)
            except Exception:
                source = ""

        first_lineno = getattr(self.ast, "lineno", 2)
        line_offset = offset(source.splitlines())

        newlines = "\n" * (first_lineno - line_offset)
        source = newlines + source

        return source.splitlines()

def offset(source_lines):
    offset = 0
    for line in source_lines:
        if line.strip().startswith("def"):
            break

        offset += 1

    return offset

# ______________________________________________________________________

def sort_message(collected_message):
    node, is_error, message = collected_message
    lineno, colno = float('inf'), float('inf')
    if hasattr(node, 'lineno'):
        lineno, colno = map(float, getpos(node))

    return not is_error, lineno, colno, message

class MessageCollection(object):
    """Collect error/warnings messages first then sort"""

    def __init__(self, ast=None, source_lines=None, file=None):
        # (node, is_error, message)
        self.file = file or sys.stdout
        self.ast = ast
        self.source_lines = source_lines
        self.messages = []
        self.have_errors = False

    def error(self, node, message):
        self.have_errors = True
        self.messages.append((node, True, message))

    def warning(self, node, message):
        self.messages.append((node, False, message))

    def header(self):
        pass

    def footer(self):
        pass

    def report_message(self, message, node, type):
        self.file.write(format_msg_simple(type, node, message))

    def report(self, post_mortem=False):
        self.messages.sort(key=sort_message)

        if self.messages:
            self.header()

        errors = []
        for node, is_error, message in self.messages:
            if is_error:
                errors.append((node, message))
                type = "Error"
            else:
                type = "Warning"

            self.report_message(message, node, type)

        if self.messages:
            self.footer()

        if errors and not post_mortem:
            raise error.NumbaError(*errors[0])


class FancyMessageCollection(MessageCollection):

    def header(self):
        self.file.write(" Numba Encountered Errors or Warnings ".center(80, "-") + '\n')

    def footer(self):
        self.file.write("-" * 80 + '\n')

    def report_message(self, message, node, type):
        self.file.write(format_msg(type, self.source_lines, node, message))

# ______________________________________________________________________

def format_msg(type, source_lines, node, msg):
    ret = ''
    if node and hasattr(node, 'lineno') and source_lines:
        lineno, colno = getpos(node)
        if lineno < len(source_lines):
            line = source_lines[lineno]
            ret = line + '\n' + "%s^" % ("-" * colno) + '\n'

    return ret + format_msg_simple(type, node, msg)

def format_msg_simple(type, node, message):
    "Issue a warning"
    # printing allows us to test the code
    return "%s %s%s\n" % (type, error.format_pos(node), message)
    # logger.warning("Warning %s: %s", error.format_postup(getpos(node)), message)

# ______________________________________________________________________

def report(env, exc=None):
    """
    :param function_error_env: the FunctionErrorEnvironment
    :param post_mortem: whether to enable post-mortem debugging of Numba
    :param exc: currently propagating exception
    """
    function_error_env = env.crnt.error_env
    post_mortem = function_error_env.enable_post_mortem
    if exc is not None:
        function_error_env.collection.error(exc.node, exc.msg)

    try:
        function_error_env.collection.report(post_mortem)
    except error.NumbaError as e:
        if exc is None:
            exc = e

    if exc is not None and not post_mortem:
        # Shorten traceback
        raise exc



#! /usr/bin/env python
# ______________________________________________________________________
'''test_filter2d

Test the filter2d() example from the PyCon'12 slide deck.
'''
# ______________________________________________________________________

import numpy

from numba import *
from numba.decorators import jit

import sys
import unittest

# ______________________________________________________________________

def filter2d(image, filt):
    M, N = image.shape
    Mf, Nf = filt.shape
    Mf2 = Mf // 2
    Nf2 = Nf // 2
    result = numpy.zeros_like(image)
    for i in range(Mf2, M - Mf2):
        for j in range(Nf2, N - Nf2):
            num = 0.0
            for ii in range(Mf):
                for jj in range(Nf):
                    num += (filt[Mf-1-ii, Nf-1-jj] * image[i-Mf2+ii, j-Nf2+jj])
            result[i, j] = num
    return result

# ______________________________________________________________________

class TestFilter2d(unittest.TestCase):
    def test_vectorized_filter2d(self):
        ufilter2d = jit(argtypes=[double[:,:], double[:,:]],
                                  restype=double[:,:])(filter2d)
        image = numpy.random.random((50, 50))
        filt = numpy.random.random((5, 5))
        filt /= filt.sum()
        plain_old_result = filter2d(image, filt)
        hot_new_result = ufilter2d(image, filt)
        self.assertTrue((abs(plain_old_result - hot_new_result) < 1e-9).all())

# ______________________________________________________________________

@autojit
def func():
    return numpy.empty(10)

if __name__ == "__main__":
#    func()
#    TestFilter2d('test_vectorized_filter2d').debug()
    unittest.main(*sys.argv[1:])

# ______________________________________________________________________
# End of test_filter2d.py



import unittest
import ast, inspect
import numpy as np
from numba import utils, decorators, environment, pipeline
from numba.minivect import minitypes
from numba import *

def cf_1():
    return 1 + 2

def cf_2():
    return 1 + 2 - 3 * 4 / 5 // 6 ** 7

def cf_3():
    return 0xbad & 0xbeef | 0xcafe

def cf_4():
    return True or False

def cf_5():
    return 1 != 2 and 3 < 4 and 5 > 8 / 9

M = 1
def cf_6():
    return M + 2

def cf_7():
    N = 1
    return N + 2

def cf_8():
    N = 1
    N += 2  # invalidate N
    return N + 3

def cf_9():
    i = 1
    j = 2
    k = 3  # the only constant
    for i, n in range(10):
        j += 0
    return i + k

def cf_10():
    i = j = 123
    return i + j

def cf_11(M):
    return M + 2

def cf_12(a):
    M = a
    return M + 2


class TestConstFolding(unittest.TestCase):
    env = environment.NumbaEnvironment.get_environment()

    def run_pipeline(self, func):
        func_sig = minitypes.FunctionType(minitypes.void, [])
        source = inspect.getsource(func)
        astree = ast.parse(source)
        with environment.TranslationContext(self.env, func, astree, func_sig):
            pipeline_callable = self.env.get_or_add_pipeline(
                'const_folding', pipeline.ConstFolding)
            astree = pipeline.AST3to2()(astree, self.env)
            ret_val = pipeline_callable(astree, self.env)
        return ret_val

    def iter_all(self, astree, target):
        for node in ast.walk(astree):
            if isinstance(node, target):
                yield node

    def test_cf_1(self):
        astree = self.run_pipeline(cf_1)
        print((utils.pformat_ast(astree)))
        nums = list(self.iter_all(astree, ast.Num))
        self.assertEqual(len(nums), 1)
        self.assertEqual(nums[0].n, (1 + 2))

    def test_cf_2(self):
        astree = self.run_pipeline(cf_2)
        print((utils.pformat_ast(astree)))
        nums = list(self.iter_all(astree, ast.Num))
        self.assertEqual(len(nums), 1)
        self.assertEqual(nums[0].n, (1 + 2 - 3 * 4 / 5 // 6 ** 7))

    def test_cf_3(self):
        astree = self.run_pipeline(cf_3)
        print((utils.pformat_ast(astree)))
        nums = list(self.iter_all(astree, ast.Num))
        self.assertEqual(len(nums), 1)
        self.assertEqual(nums[0].n, (0xbad & 0xbeef | 0xcafe))

    def test_cf_4(self):
        astree = self.run_pipeline(cf_4)
        print((utils.pformat_ast(astree)))
        names = list(self.iter_all(astree, ast.Name))
        self.assertEqual(len(names), 1)
        self.assertEqual(names[0].id, 'True')

    def test_cf_5(self):
        astree = self.run_pipeline(cf_5)
        print((utils.pformat_ast(astree)))
        names = list(self.iter_all(astree, ast.Name))
        self.assertEqual(len(names), 1)
        self.assertEqual(names[0].id, str(1 != 2 and 3 < 4 and 5 > 8 / 9))

    def test_cf_6(self):
        astree = self.run_pipeline(cf_6)
        print((utils.pformat_ast(astree)))
        names = list(self.iter_all(astree, ast.Name))
        nums = list(self.iter_all(astree, ast.Num))
        self.assertEqual(len(names), 0)
        self.assertEqual(len(nums), 1)
        self.assertEqual(nums[0].n, (1 + 2))

    def test_cf_7(self):
        astree = self.run_pipeline(cf_7)
        print((utils.pformat_ast(astree)))
        names = list(self.iter_all(astree, ast.Name))
        nums = list(self.iter_all(astree, ast.Num))
        # No removal of constant assignment
        self.assertEqual(len(names), 1)
        self.assertEqual(len(nums), 2)
        self.assertEqual(nums[1].n, (1 + 2))

    def test_cf_8(self):
        astree = self.run_pipeline(cf_8)
        print((utils.pformat_ast(astree)))
        names = list(self.iter_all(astree, ast.Name))
        nums = list(self.iter_all(astree, ast.Num))
        self.assertEqual(len(names), 3)
        self.assertEqual(len(nums), 3)
        for name in names:
            self.assertEqual(name.id, 'N')
        for i, num in enumerate(nums):
            self.assertEqual(num.n, i + 1)

    def test_cf_9(self):
        astree = self.run_pipeline(cf_9)
        print((utils.pformat_ast(astree)))
        names = list(self.iter_all(astree, ast.Name))
        nums = list(self.iter_all(astree, ast.Num))
        self.assertEqual(len(names), 8)
        self.assertEqual(len(nums), 6)

    def test_cf_10(self):
        astree = self.run_pipeline(cf_10)
        print((utils.pformat_ast(astree)))
        names = list(self.iter_all(astree, ast.Name))
        nums = list(self.iter_all(astree, ast.Num))
        self.assertEqual(len(names), 2)
        self.assertEqual(len(nums), 2)

    def test_cf_11(self):
        astree = self.run_pipeline(cf_11)
        print((utils.pformat_ast(astree)))
        names = list(self.iter_all(astree, ast.Name))
        nums = list(self.iter_all(astree, ast.Num))
        self.assertEqual(len(names), 2)
        self.assertEqual(len(nums), 1)
        self.assertEqual(nums[0].n, (2))
    
    def test_cf_12(self):
        astree = self.run_pipeline(cf_12)
        print((utils.pformat_ast(astree)))
        names = list(self.iter_all(astree, ast.Name))
        nums = list(self.iter_all(astree, ast.Num))
        self.assertEqual(len(names), 4)
        self.assertEqual(len(nums), 1)
        self.assertEqual(nums[0].n, (2))



if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
import re

_ptx_invalid_char = re.compile('[^a-zA-Z0-9_]')

def _fix_naming(string):
    def repl(m):
        return '_%X_' % (ord(m.group(0)))
    return _ptx_invalid_char.sub(repl, string)

def type_mangle(*types):
    return "_".join(str(t).replace(" ", "_") for t in types)

function_counter = 0

def specialized_mangle(func_name, types):
    global function_counter
    # pre = "__numba_specialized_%d_%s" % (func_name, type_mangle(*types))
    pre = "__numba_specialized_%d_%s" % (function_counter, func_name)
    function_counter += 1
    return _fix_naming(pre)




"""
>>> get_globals()
20
>>> get_locals()
Traceback (most recent call last):
    ...
NumbaError: locals() is not supported in numba functions
>>> get_sum(3)
6
>>> eval_something("'hello'")
'hello'
>>> list(enumerate_list())
[(0, 1), (1, 2), (2, 3)]
>>> max_(20) == 20
True
>>> min_(-2) == -2
True
"""

from numba import *

myglobal = 20

@autojit(backend='ast')
def get_globals():
    return globals()['myglobal']

@autojit(backend='ast')
def get_locals():
    x = 2
    return locals()['x']

@autojit(backend='ast')
def get_sum(x):
    return sum([1, 2, x])

@autojit(backend='ast')
def eval_something(s):
    return eval(s)

@autojit(backend='ast')
def enumerate_list():
    return enumerate([1, 2, 3])

@autojit(backend='ast')
def max_(x):
    return max(1, 2.0, x, 14)

@autojit(backend='ast')
def min_(x):
    return min(1, 2.0, x, 14)

if __name__ == '__main__':
    import numba
    numba.testing.testmod()



from __future__ import print_function

import sys
import unittest
import StringIO

from numba.minivect import minitypes
from numba import *

@autojit(backend='ast')
def print_(value):
    print(value)

@autojit(backend='ast', nopython=True)
def print_nopython(value):
    print("value", end=" ")
    print(value)

@autojit(backend='ast')
def print_to_stream(stream, value):
    print(value, file=stream)

@autojit(backend='ast')
def print_no_newline(stream, value):
    print(value, end=' ', file=stream)

class TestPrint(unittest.TestCase):
    def test_print(self):
        out = sys.stdout
        sys.stdout = temp_out = StringIO.StringIO()
        try:
            print_(10)
            print_(10.0)
            print_("hello!")
        finally:
            sys.stdout = out

        data = temp_out.getvalue()
        assert data == "10\n10.0\nhello!\n", repr(data)

    def test_print_stream(self):
        temp_out = StringIO.StringIO()
        print_to_stream(temp_out, 13.2)
        data = temp_out.getvalue()
        assert data == "13.2\n", repr(data)

    def test_print_no_newline(self):
        temp_out = StringIO.StringIO()
        print_no_newline(temp_out, 14.1)
        data = temp_out.getvalue()
        assert data == "14.1 ", repr(data)

if __name__ == "__main__":
    # The following isn't currently supported.  See issue #147
    #(https://github.com/numba/numba/issues/147).

    #print_nopython(10)

    unittest.main()



import time

import numba
from numba import *
import numpy as np

# Bug reported by Jeremiah L. Lowin

def timer(pyfunc, numbafunc, *args, **kwargs):
    t1 = time.time()
    pyresult = pyfunc(*args, **kwargs)
    t2 = time.time()
    print(('python function took: {0}'.format(t2-t1)))

    t3 = time.time()
    numbaresult = numbafunc(*args, **kwargs)
    t4 = time.time()
    print(('numba function took: {0}'.format(t4-t3)))
    print(('speedup: {0}x'.format(np.round((t2-t1) / (t4-t3),2))))

    assert np.allclose(pyresult, numbaresult)


def timer2(pyfunc, numbafunc, *args, **kwargs):
    t1 = time.time()
    pyresult = np.empty_like(args[0])
    pyargs = args + (pyresult,)
    pyfunc(*pyargs, **kwargs)
    t2 = time.time()
    print(('python function took: {0}'.format(t2-t1)))

    t3 = time.time()
    numbaresult = np.empty_like(args[0])
    nbargs = args + (numbaresult,)
    numbafunc(*nbargs, **kwargs)
    t4 = time.time()
    print(('numba function took: {0}'.format(t4-t3)))
    print(('speedup: {0}x'.format(np.round((t2-t1) / (t4-t3),2))))

    assert np.allclose(pyresult, numbaresult)


def vmdot(x, w):
    out = np.empty((x.shape[0], w.shape[1]))
    for i in range(x.shape[0]):
        dot_prod = np.dot(x[i], w)
        out[i] = np.exp(-1 * dot_prod)

    return out

def vmdot2(x, w, out):
    for i in range(x.shape[0]):
        dot_prod = np.dot(x[i], w)
        out[i] = np.exp(-1 * dot_prod)


def test_vmdot():
    numba_vmdot = jit(restype=double[:,:],
                      argtypes=[double[:,:], double[:,:]])(vmdot)
    x = np.random.random((1000, 1000))
    w = np.random.random((1000, 1000)) / 1000.
    timer(vmdot, numba_vmdot, x, w)

    # ensure this compiles
    numba_vmdot2 = jit(argtypes=[double[:,:], double[:,:], double[:,:]])(vmdot2)

    timer2(vmdot2, numba_vmdot2, x, w)

if __name__ == '__main__':
    test_vmdot()



"""
Test Python- and Numba-level inheritance.
"""

import numba
from numba import *
from numba.testing.test_support import parametrize, main

if not numba.PY3:
    # The operation is valid in Python 3

    __doc__ = """
    >>> Base.py_method(object())
    Traceback (most recent call last):
    ...
    TypeError: unbound method numba_function_or_method object must be called with Base instance as first argument (got object instance instead)
    """

def format_str(msg, *values):
    return msg % values

def make_base(compiler):
    @compiler
    class BaseClass(object):

        @void(double)
        def __init__(self, value):
            self.value = value

        @double()
        def getvalue(self):
            "Return value"
            return self.value

        @void(double)
        def setvalue(self, value):
            "Set value"
            self.value = value

        @double()
        def method(self):
            return self.getvalue()

        @double()
        def py_method(self):
            return self.value

        @object_()
        def __repr__(self):
            return format_str('Base(%s)', self.value)

    return BaseClass

def make_derived(compiler):
    BaseClass = make_base(compiler)

    @compiler
    class DerivedClass(BaseClass):

        @void(double)
        def __init__(self, value):
            self.value = value
            self.value2 = 2.0

        @double()
        def getvalue(self):
            "Return value"
            return self.value * self.value2

        @void(double)
        def setvalue2(self, value2):
            "Set value"
            self.value2 = value2

        @object_()
        def __repr__(self):
            return format_str('Derived(%s)', self.value)

    return BaseClass, DerivedClass

#------------------------------------------------------------------------
# Tests
#------------------------------------------------------------------------

@parametrize(jit, autojit)
def test_baseclass(compiler):
    Base = make_base(compiler)

    assert str(Base(10.0)) == 'Base(10.0)'
    assert Base(10.0).py_method() == 10.0

    assert Base(4.0).method() == 4.0
    assert Base(4.0).getvalue() == 4.0

    try:
        Base.py_method(object())
    except TypeError as e:
        assert e.args[0] == ('unbound method numba_function_or_method '
                             'object must be called with BaseClass '
                             'instance as first argument (got object '
                             'instance instead)'), e.args[0]
    else:
        raise Exception("Expected an exception")

@parametrize(jit) #, autojit)
def test_derivedclass(compiler):
    Base, Derived = make_derived(compiler)

    assert str(Derived(20.0)) == 'Derived(20.0)'
    assert Derived(10.0).py_method() == 10.0

    assert Derived(4.0).method() == 8.0
    assert Derived(4.0).getvalue() == 8.0

    obj = Derived(4.0)
    obj.value2 = 3.0
    assert obj.method() == 12.0


if __name__ == '__main__':
    test_baseclass(autojit)
    # test_derivedclass(autojit)
    main()



# -*- coding: utf-8 -*-

"""
Test binding of autojit methods.
"""

from __future__ import print_function, division, absolute_import

from numba import *

class A(object):
    @autojit
    def a(self, arg):
        return self * arg

    def __mul__(self, other):
        return 10 * other

assert A().a(10) == 100



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
from numba.nodes import *

class ObjectInjectNode(ExprNode):
    """
    Refer to a Python object in the llvm code.
    """

    _attributes = ['object', 'type']

    def __init__(self, object, type=None, **kwargs):
        super(ObjectInjectNode, self).__init__(**kwargs)
        self.object = object
        self.type = type or object_
        self.variable = Variable(self.type, is_constant=True,
                                 constant_value=object)

    def __repr__(self):
        return "<inject(%s)>" % self.object

NoneNode = ObjectInjectNode(None, object_)

class ObjectTempNode(ExprNode):
    """
    Coerce a node to a temporary which is reference counted.
    """

    _fields = ['node']

    def __init__(self, node, incref=False):
        assert not isinstance(node, ObjectTempNode)
        self.node = node
        self.llvm_temp = None
        self.type = getattr(node, 'type', node.variable.type)
        self.incref = incref

    def __repr__(self):
        return "objtemp(%s)" % self.node

class NoneNode(ExprNode):
    """
    Return None.
    """

    type = typesystem.NoneType()
    variable = Variable(type)

class ObjectTempRefNode(ExprNode):
    """
    Reference an ObjectTempNode, without evaluating its subexpressions.
    The ObjectTempNode must already have been evaluated.
    """

    _fields = []

    def __init__(self, obj_temp_node, **kwargs):
        super(ObjectTempRefNode, self).__init__(**kwargs)
        self.obj_temp_node = obj_temp_node


class IncrefNode(ExprNode):

    _fields = ['value']

    def __init__(self, value, **kwargs):
        super(IncrefNode, self).__init__(**kwargs)
        self.value = value

class DecrefNode(IncrefNode):
    pass



# -*- coding: utf-8 -*-
# from __future__ import division, absolute_import

# Thanks to Neal Becker

import numpy as np
from numba import *
from numba.vectorize import vectorize
from math import exp, log1p


@vectorize([f8(f8,f8)])
def log_exp_sum2 (a, b):
    if a >= b:
        return a + (exp (-(a-b)))
    else:
        return b + (exp (-(b-a)))
    ## return max (a, b) + log1p (exp (-abs (a - b)))


#@autojit
@jit(f8[:,:] (f8[:,:]))
def log_exp_sum (u):
    s = u.shape[1] # Test wraparound when implemented!
    if s == 1:
        return u[...,0]
    elif s == 2:
        return log_exp_sum2 (u[...,0], u[...,1])
    else:
        return log_exp_sum2 (
            log_exp_sum (u[...,:s/2]),
            log_exp_sum (u[...,s/2:]))


from timeit import timeit
L = 1000
N = 100
u = np.tile (np.log (np.ones (L)/L), (N, 1))
#v = log_exp_sum (u)
from timeit import timeit
print(timeit(
    'log_exp_sum(u)', 'from __main__ import u, log_exp_sum', number=50))



"""
Adpated from http://wiki.cython.org/FAQ#HowcanIrundoctestsinCythoncode.28pyxfiles.29.3F
Use the testmod function from test_support, don't use this directly.
====================================================================

Cython-compatible wrapper for doctest.testmod().

Usage example, assuming a Cython module mymod.pyx is compiled.
This is run from the command line, passing a command to Python:
python -c "import cydoctest, mymod; cydoctest.testmod(mymod)"

(This still won't let a Cython module run its own doctests
when called with "python mymod.py", but it's pretty close.
Further options can be passed to testmod() as desired, e.g.
verbose=True.)
"""
import sys
import unittest
import doctest
import inspect

import numba.decorators
from numba import numbawrapper

doctest_options = doctest.ELLIPSIS|doctest.NORMALIZE_WHITESPACE


def from_module(module, object):
    """
    Return true if the given object is defined in the given module.
    """
    if module is None:
        return True
    elif inspect.getmodule(object) is not None:
        return module is inspect.getmodule(object)
    elif inspect.isfunction(object):
        return module.__dict__ is object.__globals__
    elif inspect.isclass(object):
        return module.__name__ == object.__module__
    elif hasattr(object, '__module__'):
        return module.__name__ == object.__module__
    elif isinstance(object, property):
        return True # [XX] no way not be sure.
    else:
        raise ValueError("object must be a class or function")

def fix_module_doctest(module):
    """
    Extract docstrings from cython functions, that would be skipped by doctest
    otherwise.
    """
    module.__test__ = {}
    for name in dir(module):
        value = getattr(module, name)
        if (isinstance(value, numbawrapper.NumbaWrapper) and
                from_module(module, value.py_func) and value.py_func.__doc__):
            module.__test__[name] = value.py_func.__doc__
        elif (inspect.isbuiltin(value) and isinstance(value.__doc__, str) and
                  from_module(module, value)):
            module.__test__[name] = value.__doc__

class MyDocTestFinder(doctest.DocTestFinder):
    def find(self, obj, **kws):
        res = doctest.DocTestFinder.find(self, obj, **kws)
        return res

def testmod(m=None, run_doctests=True, optionflags=doctest_options,
            verbosity=2):
    """
    Fix a Cython module's doctests, then call doctest.testmod()
    """
    fix_module_doctest(m)
    if run_doctests:
        finder = MyDocTestFinder(exclude_empty=False)
        suite = doctest.DocTestSuite(m, test_finder=finder,
                                     optionflags=optionflags)
        result =  unittest.TextTestRunner(verbosity=verbosity).run(suite)
        if not result.wasSuccessful():
            raise Exception("Doctests failed: %s" % result)



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
from .external import ExternalFunction
from numba import *

class printf(ExternalFunction):
    arg_types = [void.pointer()]
    return_type = int32
    is_vararg = True

class puts(ExternalFunction):
    arg_types = [c_string_type]
    return_type = int32

class labs(ExternalFunction):
    arg_types = [long_]
    return_type = long_

class llabs(ExternalFunction):
    arg_types = [longlong]
    return_type = longlong

class atoi(ExternalFunction):
    arg_types = [c_string_type]
    return_type = int_

class atol(ExternalFunction):
    arg_types = [c_string_type]
    return_type = long_

class atoll(ExternalFunction):
    arg_types = [c_string_type]
    return_type = longlong

class atof(ExternalFunction):
    arg_types = [c_string_type]
    return_type = double

class strlen(ExternalFunction):
    arg_types = [c_string_type]
    return_type = size_t

__all__ = [k for k, v in globals().items()
           if (v != ExternalFunction
               and isinstance(v, type)
               and issubclass(v, ExternalFunction))]




# Based on cython/tests/run/listcomp.pyx

from numba import *
from numba.testing.test_support import testmod, autojit_py3doc

@autojit_py3doc
def smoketest():
    """
    >>> smoketest()
    ([0L, 4L, 8L], 4L)
    """
    x = -10 # 'abc'
    result = [x*2 for x in range(5) if x % 2 == 0]
    # assert x != -10 # 'abc'
    return result, x

@autojit_py3doc
def list_genexp():
    """
    >>> list_genexp()
    Traceback (most recent call last):
        ...
    NumbaError: 26:18: Generator comprehensions are not yet supported
    """
    x = -10 # 'abc'
    result = list(x*2 for x in range(5) if x % 2 == 0)
    # assert x == 'abc'
    return result, x

@autojit_py3doc(locals={"x": int_})
def int_runvar():
    """
    >>> int_runvar()
    [0L, 4L, 8L]
    """
    print([x*2 for x in range(5) if x % 2 == 0])

#@jit
#class A(object):
#    @object_()
#    def __repr__(self):
#        return u"A"

#@autojit
#def typed():
#    """
#    >>> typed()
#    [A, A, A]
#    """
#    cdef A obj
#    print [obj for obj in [A(), A(), A()]]

#@autojit
#def iterdict():
#    """
#    >>> iterdict()
#    [1, 2, 3]
#    """
#    d = dict(a=1,b=2,c=3)
#    l = [d[key] for key in d]
#    l.sort()
#    print l

@autojit_py3doc
def nested_result():
    """
    >>> nested_result()
    [[], [-1L], [-1L, 0L], [-1L, 0L, 1L]]
    """
    result = [[a-1 for a in range(b)] for b in range(4)]
    return result

# TODO: object and string iteration
#@autojit_py3doc
#def listcomp_as_condition(sequence):
#    """
#    >>> listcomp_as_condition(['a', 'b', '+'])
#    True
#    >>> listcomp_as_condition('ab+')
#    True
#    >>> listcomp_as_condition('abc')
#    False
#    """
#    if [1 for c in sequence if c in '+-*/<=>!%&|([^~,']:
#        return True
#    return False

#@autojit_py3doc
#def sorted_listcomp(sequence):
#    """
#    >>> sorted_listcomp([3,2,4])
#    [3, 4, 5]
#    """
#    return sorted([ n+1 for n in sequence ])

@autojit_py3doc
def listcomp_const_condition_false():
    """
    >>> listcomp_const_condition_false()
    []
    """
    return [x*2 for x in range(3) if 0]

@autojit_py3doc
def listcomp_const_condition_true():
    """
    >>> listcomp_const_condition_true()
    [0L, 2L, 4L]
    """
    return [x*2 for x in range(3) if 1]


if __name__ == '__main__':
#    print test_pointer_arithmetic()
#    a = np.array([1, 2, 3, 4], dtype=np.float32)
#    print test_pointer_indexing(a.ctypes.data, float32.pointer())
    pass

#int_runvar()
#smoketest()
#list_genexp()
#test_listcomp()
# smoketest()
testmod()



from numba import *
import unittest

class TestRedefine(unittest.TestCase):
    def test_redefine(self):

        def foo(x):
            return x + 1

        jfoo = jit(int32(int32))(foo)

        # Test original function
        self.assertTrue(jfoo(1), 2)


        jfoo = jit(int32(int32))(foo)

        # Test re-compiliation
        self.assertTrue(jfoo(2), 3)

        def foo(x):
            return x + 2

        jfoo = jit(int32(int32))(foo)

        # Test redefinition
        self.assertTrue(jfoo(1), 3)



from numba import *
from numba.testing.test_support import parametrize, main

def make_base(compiler):
    @compiler
    class Base(object):

        @void(double)
        def __init__(self, myfloat):
            self.value = myfloat

        @double()
        def getvalue(self):
            return self.value

        @staticmethod
        @double()
        def static1():
            return 10.0

        @double(double)
        @staticmethod
        def static2(value):
            return value * 2

        @double(double, double)
        @staticmethod
        def static3(value1, value2):
            return value1 * value2

        @staticmethod
        @double(double, double)
        def static4(value1, value2):
            return value1 * value2

        @classmethod
        @double()
        def class1(cls):
            return 10.0

        @double(double)
        @classmethod
        def class2(cls, value):
            return value * 2

        @double(double, double)
        @classmethod
        def class3(cls, value1, value2):
            return value1 * value2

        @classmethod
        @double(double, double)
        def class4(cls, value1, value2):
            return value1 * value2

    return Base

def make_derived(compiler):
    Base = make_base(compiler)

    @compiler
    class Derived(Base):
        pass

    return Base, Derived

#------------------------------------------------------------------------
# Tests
#------------------------------------------------------------------------

def run_staticmethods(obj):
    assert obj.static1() == 10.0
    assert obj.static2(10.0) == 20.0
    assert obj.static3(5.0, 6.0) == 30.0
    assert obj.static4(5.0, 6.0) == 30.0

def run_classmethods(obj):
    assert obj.class1() == 10.0
    assert obj.class2(10.0) == 20.0
    assert obj.class3(5.0, 6.0) == 30.0
    assert obj.class4(5.0, 6.0) == 30.0

# ______________________________________________________________________
# Parameterized tests

@parametrize(jit, autojit)
def test_base_staticmethods(compiler):
    Base = make_base(compiler)
    run_staticmethods(Base(2.0))
    run_staticmethods(Base)

@parametrize(jit)
def test_derived_staticmethods(compiler):
    Base, Derived = make_derived(compiler)
    run_staticmethods(Derived(2.0))
    run_staticmethods(Derived)

@parametrize(jit, autojit)
def test_base_classmethods(compiler):
    Base = make_base(compiler)
    run_classmethods(Base(2.0))
    run_classmethods(Base)

@parametrize(jit)
def test_derived_classmethods(compiler):
    Base, Derived = make_derived(compiler)
    run_classmethods(Derived(2.0))
    run_classmethods(Derived)

if __name__ == '__main__':
    # Base = make_base(autojit)
    # obj = Base(2.0)
    # run_staticmethods(Base)
    main()



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
from numba.nodes import *
from numba import functions

class LLVMValueRefNode(ExprNode):
    """
    Wrap an LLVM value.
    """

    _fields = []

    def __init__(self, type, llvm_value):
        self.type = type
        self.llvm_value = llvm_value

class BadValue(LLVMValueRefNode):
    def __init__(self, type):
        super(BadValue, self).__init__(type, None)

    def __repr__(self):
        return "bad(%s)" % self.type

class LLVMCBuilderNode(UserNode):
    """
    Instantiate an link in an LLVM cbuilder CDefinition. The CDefinition is
    passed the list of dependence nodes and the list of LLVM value dependencies
    """

    _fields = ["dependencies"]

    def __init__(self, env, cbuilder_cdefinition, signature, dependencies=None):
        self.env = env
        self.llvm_context = env.llvm_context
        self.cbuilder_cdefinition = cbuilder_cdefinition
        self.type = signature
        self.dependencies = dependencies or []

    def infer_types(self, type_inferer):
        type_inferer.visitchildren(self)
        return self

    def codegen(self, codegen):
        func_env = self.env.translation.crnt

        dependencies = codegen.visitlist(self.dependencies)
        cdef = self.cbuilder_cdefinition(self.dependencies, dependencies)

        self.lfunc = cdef.define(func_env.llvm_module) #, optimize=False)
        functions.keep_alive(func_env.func, self.lfunc)

        return self.lfunc

    @property
    def pointer(self):
        return self.llvm_context.get_pointer_to_function(self.lfunc)



# -*- coding: utf-8 -*-

"""
Virtual methods using virtual method tables.

Note that for @jit classes, we do not support multiple inheritance with
incompatible base objects. We could use a dynamic offset to base classes,
and adjust object pointers for method calls, like in C++:

    http://www.phpcompiler.org/articles/virtualinheritance.html

However, this is quite complicated, and still doesn't allow dynamic extension
for autojit classes. Instead we will use Dag Sverre Seljebotn's hash-based
virtual method tables:

    https://github.com/numfocus/sep/blob/master/sep200.rst
    https://github.com/numfocus/sep/blob/master/sep201.rst
"""

import ctypes

import numba
from numba.typesystem import *
from numba.minivect import minitypes
from numba.exttypes import ordering
from numba.exttypes import extension_types

#------------------------------------------------------------------------
# Virtual Method Table Interface
#------------------------------------------------------------------------

class VTabBuilder(object):
    """
    Build virtual method table for quick calling from Numba.
    """

    def finalize(self, ext_type):
        "Finalize the method table (and fix the order if necessary)"

    def build_vtab(self, ext_type, method_pointers):
        """
        Build a virtual method table.
        The result will be kept alive on the extension type.
        """

    def wrap_vtable(self, vtable):
        """
        Wrap the vtable such that users can get a pointer to the underlying
        data (extension_types.{Static,Dynamic}VtableWrapper).
        """

#------------------------------------------------------------------------
# Static Virtual Method Tables
#------------------------------------------------------------------------

def vtab_name(field_name):
    "Mangle method names for the vtab (ctypes doesn't handle this)"
    if field_name.startswith("__") and field_name.endswith("__"):
        field_name = '__numba_' + field_name.strip("_")
    return field_name

def build_static_vtab(vtable, vtab_struct):
    """
    Create ctypes virtual method table.

    vtab_type: the vtab struct type (typesystem.struct)
    method_pointers: a list of method pointers ([int])
    """
    vtab_ctype = numba.struct(
        [(vtab_name(field_name), field_type)
            for field_name, field_type in vtab_struct.fields]).to_ctypes()

    methods = []
    for method, (field_name, field_type) in zip(vtable.methods,
                                                vtab_struct.fields):
        method_type_p = field_type.to_ctypes()
        method_void_p = ctypes.c_void_p(method.lfunc_pointer)
        cmethod = ctypes.cast(method_void_p, method_type_p)
        methods.append(cmethod)

    vtab = vtab_ctype(*methods)
    return vtab

# ______________________________________________________________________
# Build Virtual Method Table

class StaticVTabBuilder(VTabBuilder):

    def finalize(self, ext_type):
        ext_type.vtab_type.create_method_ordering(ordering.extending)

    def build_vtab(self, ext_type):
        vtable = ext_type.vtab_type
        return build_static_vtab(vtable, vtable.to_struct())

    def wrap_vtable(self, vtable):
        return extension_types.StaticVtableWrapper(vtable)

#------------------------------------------------------------------------
# Hash-based virtual method tables
#------------------------------------------------------------------------

# ______________________________________________________________________
# Type Definitions

# TODO: Use something like CFFI + type conversion to get these
# TODO: types automatically

PyCustomSlots_Entry = numba.struct([
    ('id', uint64),
    ('ptr', void.pointer()),
])

PyCustomSlots_Table = numba.struct([
    ('flags', uint64),
    ('m_f', uint64),
    ('m_g', uint64),
    ('entries', PyCustomSlots_Entry.pointer()),
    ('n', uint16),
    ('b', uint16),
    ('r', uint8),
    ('reserved', uint8),
    # actually: uint16[b], 'b' trailing displacements
    # ('d', minitypes.CArrayType(uint16, 0)), #0xffff)),
    # ('entries_mem', PyCustomSlot_Entry[n]), # 'n' trailing customslot entries
])

# ______________________________________________________________________
# Hash-table building

def initialize_interner():
    from numba.pyextensibletype.extensibletype import intern
    intern.global_intern_initialize()

def sep201_signature_string(functype, name):
    functype = minitypes.FunctionType(functype.return_type, functype.args, name)
    return str(functype)

def hash_signature(functype, name):
    from numba.pyextensibletype.extensibletype import methodtable

    initialize_interner()
    sep201_hasher = methodtable.Hasher()

    sigstr = sep201_signature_string(functype, name)
    return sep201_hasher.hash_signature(sigstr)

def build_hashing_vtab(vtable):
    """
    Build hash-based vtable.
    """
    from numba.pyextensibletype.extensibletype import methodtable

    n = len(vtable.methods)

    ids = [sep201_signature_string(method.signature, method.name)
               for method in vtable.methods]
    flags = [0] * n

    sep201_hasher = methodtable.Hasher()
    vtab = methodtable.PerfectHashMethodTable(sep201_hasher)
    vtab.generate_table(n, ids, flags, list(vtable.method_pointers))
    # print(vtab)
    return vtab

# ______________________________________________________________________
# Build Hash-based Virtual Method Table

class HashBasedVTabBuilder(VTabBuilder):

    def finalize(self, ext_type):
        ext_type.vtab_type.create_method_ordering(ordering.unordered)

    def build_vtab(self, ext_type):
        return build_hashing_vtab(ext_type.vtab_type)

    def wrap_vtable(self, vtable):
        return extension_types.DynamicVtableWrapper(vtable)



from itertools import product
from numba import autojit

def _make_test(f):
    def test_ifexp():
        f_ = autojit(f)
        for args in product(range(3), range(3)):
            assert f_(*args)==f(*args)
    test_ifexp.__name__ = f.__name__
    return test_ifexp

@_make_test
def test_as_return_value(a,b):
    return a if a>b else b

@_make_test
def test_assign_and_return(a,b):
    c = a if a>b else b
    return c

@_make_test
def test_in_expression(a,b):
    c = 5 + (a if a>b else b)/2.0
    return c

@_make_test
def test_expr_as_then_clause(a,b):
    return (a+1) if a>b else b

@_make_test
def test_expr_as_else_clause(a,b):
    return a if a>b else (b+1)


@autojit
def _f1(a,b):
    return a if a>b else b

def test_type_promotion():
    assert isinstance(_f1(1, 1), long)
    assert isinstance(_f1(1.0, 1), float)
    assert isinstance(_f1(1, 1.0), float)

if __name__ == '__main__':
    test_type_promotion()



import sys
import numba

def test_nosource():
    source = '''
@numba.autojit
def foo (): return 99

'''
    new_locals = {}
    exec(source, globals(), new_locals)
    foo = new_locals['foo']
    assert foo() == 99

if sys.version_info[:2] < (2, 7):
    del test_nosource
elif __name__ == "__main__":
    test_nosource()



import sys

import math
import numpy as np
import unittest
#import logging; logging.getLogger().setLevel(1)

from numba import *

def exp_fn(a):
    return math.exp(a)

def sqrt_fn(a):
    return math.sqrt(a)

def log_fn(a):
    return math.log(a)

class TestNoPythonMath(unittest.TestCase):
    def test_sqrt(self):
        self._template(sqrt_fn, np.sqrt)

    def test_exp(self):
        self._template(exp_fn, np.exp)

    def test_log(self):
        self._template(log_fn, np.log)

    def _template(self, func, npfunc):
        func_jitted = jit(argtypes=[f4], restype=f4, nopython=True)(func)
        A = np.array(np.random.random(10), dtype=np.float32)
        B = np.vectorize(func_jitted)(A)
        self.assertTrue(np.allclose(B, npfunc(A)))

if sys.platform == 'win32':
    # NOTE: we're using the double implementation (e.g. 'log' instead of 'logf')
    # class TestNoPythonMath(unittest.TestCase):
    #     """
    #     LLVM intrinsics don't work properly on Windows, and libc doesn't
    #     have all these functions.
    #     """
    pass

if __name__ == '__main__':
    unittest.main()



from numba import *
from numba import vectorize, typeof

import numpy as np

@vectorize.vectorize([
    bool_(double, int_),
    double(double, double),
    float_(double, float_),
])
def add(a, b):
    return a + b

@autojit
def func(dtypeA, dtypeB):
    A = np.arange(10, dtype=dtypeA)
    B = np.arange(10, dtype=dtypeB)
    return typeof(add(A, B))

if __name__ == "__main__":
    assert func(np.dtype(np.float64), np.dtype('i')) == int8[:]
    assert func(np.dtype(np.float64), np.dtype(np.float64)) == double[:]
    assert func(np.dtype(np.float64), np.dtype(np.float32)) == float_[:]



#! /usr/bin/env python
# ______________________________________________________________________
'''test_forloop

Test the Numba compiler on a simple for loop over an iterable object.
'''
# ______________________________________________________________________

from numba import *
from numba.testing import test_support

import numpy

import unittest
try:
    import __builtin__ as builtins
except ImportError:
    import builtins

# ______________________________________________________________________

def for_loop_fn_0 (iterable):
    acc = 0.
    for value in iterable:
        acc += value
    return acc

# ______________________________________________________________________

def for_loop_fn_1 (start, stop, inc):
    acc = 0
    for value in range(start, stop, inc):
        acc += value
    return acc

# ______________________________________________________________________

def for_loop_fn_2 (stop):
    acc = 0
    for value_0 in range(stop):
        for value_1 in range(stop):
            acc += value_0 * value_1
    return acc

# ______________________________________________________________________

def for_loop_fn_3 (stop):
    acc = 0
    for i in range(stop):
        for j in range(stop):
            for k in range(stop):
                for l in range(stop):
                    acc += 1
    return acc

# ______________________________________________________________________

def for_loop_w_guard_0 (test_input):
    '''Test case based on issue #25.  See:
    https://github.com/numba/numba/issues/25'''
    acc = 0.0
    for i in range(5):
        if i == test_input:
            acc += 100.0
    return acc

# ______________________________________________________________________

def for_loop_w_guard_1 (test_input):
    '''Test case based on issue #25.  See:
    https://github.com/numba/numba/issues/25'''
    acc = 0.0
    for i in range(5):
        if i == test_input:
            acc += 100.0
        else:
            acc += i
    return acc

# ______________________________________________________________________

def for_loop_fn_4(i, u, p, U):
    '''Test case for issue #48.  See:
    https://github.com/numba/numba/issues/48'''
    s = 0
    t = 0
    for j in range(-p, p+2):
        if U[i+j] == u:
            t = t + 1
        if u == U[i+j]:
            s = s + 1
    if t != s:
        s = -1
    return s

# ______________________________________________________________________

class TestForLoop(unittest.TestCase):
    @test_support.skip_unless(hasattr(builtins, '__noskip__'),
                              "Requires implementation of iteration "
                              "over arrays.")
    def test_compiled_for_loop_fn_0(self):
        test_data = numpy.array([1, 2, 3], dtype = 'l')
        compiled_for_loop_fn = jit(restype=f4,
            argtypes = [i8[:]],backend='ast')(for_loop_fn_0)
        result = compiled_for_loop_fn(test_data)
        self.assertEqual(result, 6)
        self.assertEqual(result, for_loop_fn_0(test_data))

    def test_compiled_for_loop_fn_1(self):
        compiled_for_loop_fn = jit(argtypes = [i4, i4, i4],
                                             restype = i4, backend='ast')(for_loop_fn_1)
        result = compiled_for_loop_fn(1, 4, 1)
        self.assertEqual(result, 6)
        self.assertEqual(result, for_loop_fn_1(1, 4, 1))

    def test_compiled_for_loop_fn_2(self):
        compiled_for_loop_fn = jit(argtypes = [i4],
                                             restype = i4, backend='ast')(for_loop_fn_2)
        result = compiled_for_loop_fn(4)
        self.assertEqual(result, 36)
        self.assertEqual(result, for_loop_fn_2(4))

    def test_compiled_for_loop_fn_3(self):
        compiled_for_loop_fn = jit(argtypes = [i4],
                                             restype = i4, backend='ast')(for_loop_fn_3)
        result = compiled_for_loop_fn(3)
        self.assertEqual(result, for_loop_fn_3(3))
        self.assertEqual(result, 81)

    def test_compiled_for_loop_w_guard_0(self):
        compiled_for_loop_w_guard = autojit(backend='ast')(for_loop_w_guard_0)
        self.assertEqual(compiled_for_loop_w_guard(5.),
                         for_loop_w_guard_0(5.))
        self.assertEqual(compiled_for_loop_w_guard(4.),
                         for_loop_w_guard_0(4.))

    def test_compiled_for_loop_w_guard_1(self):
        compiled_for_loop_w_guard = autojit(backend='ast')(for_loop_w_guard_1)
        self.assertEqual(compiled_for_loop_w_guard(5.),
                         for_loop_w_guard_1(5.))
        self.assertEqual(compiled_for_loop_w_guard(4.),
                         for_loop_w_guard_1(4.))

    def test_compiled_for_loop_fn_4(self):
        compiled = jit('i4(i4,f8,i4,f8[:])')(for_loop_fn_4)
        args0 = 5, 1.0, 2, numpy.ones(10)
        self.assertEqual(compiled(*args0), for_loop_fn_4(*args0))
        args1 = 5, 1.0, 2, numpy.zeros(10)
        self.assertEqual(compiled(*args1), for_loop_fn_4(*args1))

# ______________________________________________________________________

if __name__ == "__main__":
#    compiled_for_loop_fn = jit(argtypes = [i4, i4, i4],
#                               restype = i4, backend='ast', nopython=True)(for_loop_fn_1)
#    result = compiled_for_loop_fn(1, 4, 1)
#    compiled_for_loop_fn = jit(argtypes = [i4],
#                               restype = i4, backend='ast')(for_loop_fn_3)
#    result = compiled_for_loop_fn(3)
#    compiled = jit('i4(i4,f8,i4,f8[:])')(for_loop_fn_4)
#    args0 = 5, 1.0, 2, numpy.ones(10)
#    print compiled(*args0)
#    print for_loop_fn_4(*args0)
    unittest.main()

# ______________________________________________________________________
# End of test_forloop.py



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

# Import all special functions before registering the Numba module
# type inferer
from numba.special import *

import os
import sys
import logging

from numba import utils, typesystem

PY3 = sys.version_info[0] == 3

def get_include():
    numba_root = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(numba_root, "include")

# NOTE: Be sure to keep the logging level commented out before commiting.  See:
#   https://github.com/numba/numba/issues/31
# A good work around is to make your tests handle a debug flag, per
# numba.tests.test_support.main().

class _RedirectingHandler(logging.Handler):
    '''
    A log hanlder that applies its formatter and redirect the emission
    to a parent handler.
    '''
    def set_handler(self, handler):
        self.handler = handler

    def emit(self, record):
        # apply our own formatting
        record.msg = self.format(record)
        record.args = [] # clear the args
        # use parent handler to emit record
        self.handler.emit(record)

def _config_logger():
    root = logging.getLogger(__name__)
    format = "\n\033[1m%(levelname)s -- "\
             "%(module)s:%(lineno)d:%(funcName)s\033[0m\n%(message)s"
    try:
        parent_hldr = root.parent.handlers[0]
    except IndexError: # parent handler is not initialized?
        # build our own handler --- uses sys.stderr by default.
        parent_hldr = logging.StreamHandler()
    hldr = _RedirectingHandler()
    hldr.set_handler(parent_hldr)
    fmt = logging.Formatter(format)
    hldr.setFormatter(fmt)
    root.addHandler(hldr)
    root.propagate = False # do not propagate to the root logger

_config_logger()


from . import special
from numba.typesystem import *
from numba.minivect.minitypes import FunctionType
from numba.error import *

from numba.containers.typedlist import typedlist
from numba.containers.typedtuple import typedtuple
from numba.typesystem import map_dtype
from numba.type_inference.module_type_inference import (is_registered,
                                                        register,
                                                        register_inferer,
                                                        get_inferer,
                                                        register_unbound,
                                                        register_callable)
from numba.typesystem.typeset import *

from numba.codegen import translate
from numba.decorators import *
from numba import decorators
from numba.intrinsic.numba_intrinsic import (declare_intrinsic,
                                             declare_instruction)

__all__ = typesystem.__all__ + decorators.__all__ + special.__all__
__all__.extend(["numeric", "floating", "complextypes"])

from numba import testing
from numba.testing import test, nose_run
from numba.testing.user_support import testmod



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
import unittest

from numba.asdl import schema

def build_schema():
    '''Build a schema from Python.asdl
    '''
    return schema.load('Python.asdl')

class SchemaTestCase(unittest.TestCase):
    '''A base class for test cases that use the Python.asdl
    '''
    schema = build_schema()

    def capture_error(self):
        return self.assertRaises(schema.SchemaError)



def _groupn(iter, N):
    k = 0
    ret = ()
    for item in iter:
        k = k + 1
        ret += (item,)
        if k == N:
            yield ret
            k = 0
            ret = ()

_toreplace = [('unsigned long ', 'ct.c_ulong '),
              ('long ', 'ct.c_long '),
              ('double ', 'ct.c_double '),
              ('rk_state *', 'rk_address '),
              ('void *', 'ct.c_void_p '),
              ('void', 'None')]
        
# Return [name, restype, and argtypes, argnames]
def parse_header(header):
    mystr = '@@@_%d_@@@'
    for i, (old, new) in enumerate(_toreplace):
        header = header.replace(old, mystr % i)
    for i, (old, new) in enumerate(_toreplace):
        header = header.replace(mystr % i, new)
    funcs = [val.split(';')[0].strip() for indx, val in enumerate(header.split('extern')) if indx > 0]
    result = []
    for func in funcs:
        temp = func.split()
        restype = temp[0]
        name, arg0type = temp[1].split('(')
        argtypes = [arg0type]
        argnames = [temp[2].strip(')')]
        for atype, aname in _groupn(temp[3:], 2):
            argnames.append(aname.strip(')'))
            argtypes.append(atype)
        result.append([name, restype, argtypes, argnames])
    return result

header1 = """
/*
 * Initialize the RNG state using the given seed.
 */

/*
 * Initialize the RNG state using a random seed.
 * Uses /dev/random or, when unavailable, the clock (see randomkit.c).
 * Returns RK_NOERR when no errors occurs.
 * Returns RK_ENODEV when the use of RK_DEV_RANDOM failed (for example because
 * there is no such device). In this case, the RNG was initialized using the
 * clock.
 */
extern rk_error rk_randomseed(rk_state *state);

/*
 * Returns a random unsigned long between 0 and RK_MAX inclusive
 */
extern unsigned long rk_random(rk_state *state);

/*
 * Returns a random long between 0 and LONG_MAX inclusive
 */
extern long rk_long(rk_state *state);

/*
 * Returns a random unsigned long between 0 and ULONG_MAX inclusive
 */
extern unsigned long rk_ulong(rk_state *state);


/*
 * Returns a random double between 0.0 and 1.0, 1.0 excluded.
 */
extern double rk_double(rk_state *state);

/*
 * return a random gaussian deviate with variance unity and zero mean.
 */
extern double rk_gauss(rk_state *state);
"""

header2 = """ 
/* Normal distribution with mean=loc and standard deviation=scale. */
extern double rk_normal(rk_state *state, double loc, double scale);

/* Standard exponential distribution (mean=1) computed by inversion of the 
 * CDF. */
extern double rk_standard_exponential(rk_state *state);

/* Exponential distribution with mean=scale. */
extern double rk_exponential(rk_state *state, double scale);

/* Uniform distribution on interval [loc, loc+scale). */
extern double rk_uniform(rk_state *state, double loc, double scale);

/* Standard gamma distribution with shape parameter. 
 * When shape < 1, the algorithm given by (Devroye p. 304) is used.
 * When shape == 1, a Exponential variate is generated.
 * When shape > 1, the small and fast method of (Marsaglia and Tsang 2000) 
 * is used.
 */
extern double rk_standard_gamma(rk_state *state, double shape);

/* Gamma distribution with shape and scale. */
extern double rk_gamma(rk_state *state, double shape, double scale);

/* Beta distribution computed by combining two gamma variates (Devroye p. 432).
 */
extern double rk_beta(rk_state *state, double a, double b);

/* Chi^2 distribution computed by transforming a gamma variate (it being a
 * special case Gamma(df/2, 2)). */
extern double rk_chisquare(rk_state *state, double df);

/* Noncentral Chi^2 distribution computed by modifying a Chi^2 variate. */
extern double rk_noncentral_chisquare(rk_state *state, double df, double nonc);

/* F distribution computed by taking the ratio of two Chi^2 variates. */
extern double rk_f(rk_state *state, double dfnum, double dfden);

/* Noncentral F distribution computed by taking the ratio of a noncentral Chi^2
 * and a Chi^2 variate. */
extern double rk_noncentral_f(rk_state *state, double dfnum, double dfden, double nonc);

/* Binomial distribution with n Bernoulli trials with success probability p.
 * When n*p <= 30, the "Second waiting time method" given by (Devroye p. 525) is
 * used. Otherwise, the BTPE algorithm of (Kachitvichyanukul and Schmeiser 1988)
 * is used. */
extern long rk_binomial(rk_state *state, long n, double p);

/* Binomial distribution using BTPE. */
extern long rk_binomial_btpe(rk_state *state, long n, double p);

/* Binomial distribution using inversion and chop-down */
extern long rk_binomial_inversion(rk_state *state, long n, double p);

/* Negative binomial distribution computed by generating a Gamma(n, (1-p)/p)
 * variate Y and returning a Poisson(Y) variate (Devroye p. 543). */
extern long rk_negative_binomial(rk_state *state, double n, double p);

/* Poisson distribution with mean=lam.
 * When lam < 10, a basic algorithm using repeated multiplications of uniform
 * variates is used (Devroye p. 504).
 * When lam >= 10, algorithm PTRS from (Hoermann 1992) is used.
 */
extern long rk_poisson(rk_state *state, double lam);

/* Poisson distribution computed by repeated multiplication of uniform variates.
 */
extern long rk_poisson_mult(rk_state *state, double lam);

/* Poisson distribution computer by the PTRS algorithm. */
extern long rk_poisson_ptrs(rk_state *state, double lam);

/* Standard Cauchy distribution computed by dividing standard gaussians 
 * (Devroye p. 451). */
extern double rk_standard_cauchy(rk_state *state);

/* Standard t-distribution with df degrees of freedom (Devroye p. 445 as
 * corrected in the Errata). */
extern double rk_standard_t(rk_state *state, double df);

/* von Mises circular distribution with center mu and shape kappa on [-pi,pi]
 * (Devroye p. 476 as corrected in the Errata). */
extern double rk_vonmises(rk_state *state, double mu, double kappa);

/* Pareto distribution via inversion (Devroye p. 262) */
extern double rk_pareto(rk_state *state, double a);

/* Weibull distribution via inversion (Devroye p. 262) */
extern double rk_weibull(rk_state *state, double a);

/* Power distribution via inversion (Devroye p. 262) */
extern double rk_power(rk_state *state, double a);

/* Laplace distribution */
extern double rk_laplace(rk_state *state, double loc, double scale);

/* Gumbel distribution */
extern double rk_gumbel(rk_state *state, double loc, double scale);

/* Logistic distribution */
extern double rk_logistic(rk_state *state, double loc, double scale);

/* Log-normal distribution */
extern double rk_lognormal(rk_state *state, double mean, double sigma);

/* Rayleigh distribution */
extern double rk_rayleigh(rk_state *state, double mode);

/* Wald distribution */
extern double rk_wald(rk_state *state, double mean, double scale);

/* Zipf distribution */
extern long rk_zipf(rk_state *state, double a);

/* Geometric distribution */
extern long rk_geometric(rk_state *state, double p);
extern long rk_geometric_search(rk_state *state, double p);
extern long rk_geometric_inversion(rk_state *state, double p);

/* Hypergeometric distribution */
extern long rk_hypergeometric(rk_state *state, long good, long bad, long sample);
extern long rk_hypergeometric_hyp(rk_state *state, long good, long bad, long sample);
extern long rk_hypergeometric_hrua(rk_state *state, long good, long bad, long sample);

/* Triangular distribution */
extern double rk_triangular(rk_state *state, double left, double mode, double right);

/* Logarithmic series distribution */
extern long rk_logseries(rk_state *state, double p);
"""

#Create function and name
ctypes_template = """
{func_name} = mtrand.{func_name}
{func_name}.restype = {restype}
{func_name}.argtypes = {argtypes}
"""

func_template = """
def {pyfunc_name}({args}):
    return {func_name}({plus_args})
"""

_result1 = parse_header(header1)
_result2 = parse_header(header2)
_result = _result1 + _result2

afile = open('_rng_generated.py','w')

for func in _result:
    argstr = "[" + ",".join(func[2]) + "]"
    afile.write(ctypes_template.format(func_name = func[0],
                                restype = func[1], argtypes = argstr))
    args = [x.strip(',') for x in func[3][1:]]
    paramstr = ",".join(args)
    allparams = ",".join(['state_p']+args)
    afile.write(func_template.format(func_name = func[0], pyfunc_name = func[0][3:],
                                     args = paramstr, plus_args=allparams))




import unittest
from itertools import product
from numba import jit

@jit('uint32(uint32, uint32)')
def bitlshift(a, b):
    return (a << b) & 0xffffffff

@jit('uint32(uint32, uint32)')
def bitrshift(a, b):
    return (a >> b) & 0xffffffff

@jit('int32(int32, int32)')
def bitashift(a, b):
    return (a >> b) & 0xffffffff

@jit('uint32(uint32, uint32)')
def bitand(a, b):
    return a & b

@jit('uint32(uint32, uint32)')
def bitor(a, b):
    return a | b

@jit('uint32(uint32)')
def bitnot(a):
    return ~a

@jit('uint32(uint32, uint32)')
def bitxor(a, b):
    return a ^ b

def boundary_range():
    maxint = 2**32 - 1
    blimit = 10
    for i in xrange(0, blimit):
        yield i
    i = blimit
    while i < maxint:
        i &= 0xffffffff
        yield i
        i = i << 1
    for i in xrange(maxint - blimit, maxint):
        yield i

class TestBitwise(unittest.TestCase):
    def template(self, testfn, expfn, op):
        for a, b in product(boundary_range(), boundary_range()):
            exp = expfn(a, b)
            try:
                got = testfn(a, b)
            except:
                print('Exception raised at a=%s b=%s, exp=%s' % (a, b, exp))
                raise
            self.assertTrue(exp == got,
                            "%s %s %s -> %s (expect %s)" % (a, op, b, got, exp))

    def test_lshift(self):
        self.template(bitlshift, bitlshift.py_func, '<<')

    def test_rshift(self):
        self.template(bitrshift, bitrshift.py_func, '>>')

    def test_ashift(self):
        self.template(bitashift, bitashift.py_func, '>>')

    def test_and(self):
        self.template(bitand, bitand.py_func, '&')

    def test_or(self):
        self.template(bitor, bitor.py_func, '|')

    def test_xor(self):
        self.template(bitxor, bitxor.py_func, '^')


if __name__ == '__main__':
    unittest.main()



"""
Extension attribute Variables used for attribute type inference.
See also compileclass.build_extension_symtab().
"""

from numba import symtab

class ExtensionAttributeVariable(symtab.Variable):
    """
    Variable created during type inference for assignments to extension
    attributes for which we don't know the type yet.

    When the assignment happens, update ext_type.attributedict.
    """

    def __init__(self, ext_type, attr_name, type, *args, **kwargs):
        super(ExtensionAttributeVariable, self).__init__(type, *args, **kwargs)
        self.ext_type = ext_type
        self.attr_name = attr_name

    def perform_assignment(self, rhs_type):
        self.type = rhs_type
        self.ext_type.attributedict[self.attr_name] = rhs_type



"""
>>> from numba.tests.closures import test_closure_type_inference
"""

import numpy as np

from numba import *
from numba.testing.test_support import *

@autojit
def test_cellvar_promotion(a):
    """
    >>> inner = test_cellvar_promotion(10)
    200.0
    >>> inner.__name__
    'inner'
    >>> inner()
    1000.0
    """
    b = int(a) * 2

    @jit(void())
    def inner():
        print((a * b))

    inner()
    a = float(a)
    b = a * a # + 1j # Promotion issue
    return inner

@autojit
def test_cellvar_promotion_error(a):
    """
    >>> from numba.minivect import minierror
    >>> try:
    ...     test_cellvar_promotion_error(10)
    ... except minierror.UnpromotableTypeError as e:
    ...     print(sorted(e.args, key=str))
    [(int, const char *)]
    """
    b = int(a) * 2

    @jit(void())
    def inner():
        print((a * b))

    inner()
    a = np.empty(10, dtype=np.double)
    b = "hello"
    return inner

#test_cellvar_promotion(10)
#test_cellvar_promotion_error(10)
testmod()



# -*- coding: utf-8 -*-

"""
Autojit meta class.
"""

from __future__ import print_function, division, absolute_import

class _AutojitMeta(type):
    """
    Metaclass base for autojit classes.
    """

def create_unspecialized_cls(py_class, class_specializer):
    """
    Create an unspecialized class.

    class_specializer:
        NumbaSpecializingWrapper (args -> specialized object instance)
    """

    class AutojitMeta(type(py_class)):
        """
        Metaclass base for autojit classes.

            AutojitMeta -> UnspecializedClass -> SpecializedInstance
        """

        def __call__(cls, *args, **kwargs):
            return class_specializer(*args, **kwargs)

        def __getitem__(cls, key):
            assert isinstance(key, dict)

            for specialized_cls in cls.specializations:
                attrdict = specialized_cls.exttype.attribute_table.attributedict
                if attrdict == key:
                    return specialized_cls

            raise KeyError(key)

        @property
        def specializations(cls):
            return class_specializer.funccache.specializations.values()

        # def __repr__(cls):
        #     return "<Unspecialized Class %s at 0x%x>" % (cls.__name__, id(cls))


    return AutojitMeta(py_class.__name__,
                       py_class.__bases__,
                       dict(vars(py_class)))


def create_specialized_metaclass(py_class):
    """
    When the autojit cache compiles a new class specialization, it invokes
    it with the constructor arguments. Since A_specialized inherits from A,
    AutojitMeta.__call__ again tries to specialize the class. We need to
    override this behaviour and instead instantiate A_specialized through
    type.__call__ (invoking A_specialized.{__new__,__init__}).
    """

    class SpecializedMeta(type(py_class)):
        def __call__(cls, *args, **kwargs):
            return type.__call__(cls, *args, **kwargs)

        # def __repr__(cls):
        #     return "<Specialized Class %s at 0x%x>" % (cls.__name__, id(cls))

    return SpecializedMeta



# issue: #68
# Thanks to tpievila

from numba import *

import numpy as np
from numpy.random import randn
from numpy import zeros, double

nple = 64
k_fene = 15.0
R0 = 2.0
R0_2 = R0*R0

sigma = 1.0
sigma2 = sigma*sigma
sigma6 = sigma2*sigma2*sigma2
sigma12 = sigma6*sigma6

dpart = 1.12246*sigma
rcut = dpart
r2cut = dpart*dpart
k_fene = 15.0
force = 5.0 # Constant bias force in ext_force
width = 0.8775  # for ext force

@autojit()
def ext_force1(x, y, extx, exty):
    for i in xrange(nple):
        if abs(x[i]) > rcut:
            extx[i] = 0.0
        elif abs(y[i]) > rcut + width:
            r2dist = 1.0 / (x[i] * x[i])
            r6dist = r2dist * r2dist * r2dist
            extx[i] = (x[i] * 48.0 * r6dist * r2dist *
                       (sigma12 * r6dist - 0.5 * sigma6))
        else:
            ydista = y[i] - width - rcut
            ydistb = y[i] + width + rcut
            dist2a = x[i] * x[i] + ydista * ydista
            dist2b = x[i] * x[i] + ydistb * ydistb
            if dist2b < r2cut or dist2a < r2cut:
                if dist2b < r2cut:
                    ydist = ydistb
                    r2dist = 1.0 / dist2b
                else:
                    ydist = ydista
                    r2dist = 1.0 / dist2a
                r6dist = r2dist * r2dist * r2dist
                lj_factor = 48.0 * r6dist * r2dist * (sigma12 * r6dist - 0.5 * sigma6)
                extx[i] = x[i] * lj_factor
                exty[i] = ydist * lj_factor
        if abs(x[i]) < 0.5 and abs(y[i]) < width + rcut:
            extx[i] += force  # constant bias force in the x-direction
    return extx, exty


@autojit()
def ext_force2(x, y):
    extx = object_(zeros(nple, double))
    exty = object_(zeros(nple, double))
    ext_force1(x, y, extx, exty)
    return extx, exty


def test_forces():
    extx = zeros(nple, double)
    exty = zeros(nple, double)

    x, y = randn(nple), randn(nple)
    x1, y1 = ext_force1(x, y, extx, exty)
    x2, y2 = ext_force2(x, y)

    assert np.allclose(x1, x2), x2 - x1
    assert np.allclose(y1, y2), y2 - y1

if __name__ == '__main__':
    test_forces()



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

import numpy as np
from numba import autojit

@autojit
def get_jenks_breaks(data, lower_class_limits, n_classes):
    k = int(len(data) - 1)
    for countNum in range(n_classes):
        # problem inferring `k` in slice
        k = int(lower_class_limits[k, countNum] - 1)

def test():
    n_classes = 5
    data = np.ones(10)
    lower_class_limits = np.empty((data.size + 1, n_classes + 1))
    #
    #        File "/Users/sklam/dev/numba/numba/type_inference/infer.py", line 1055, in visit_Subscript
    #    if slice_type.variable.type.is_unresolved:
    #        File "/Users/sklam/dev/numba/numba/minivect/minitypes.py", line 492, in __getattr__
    #            return getattr(type(self), attr)
    #AttributeError: type object 'TupleType' has no attribute 'variable'
    print((get_jenks_breaks(data, lower_class_limits, n_classes)))

if __name__ == '__main__':
    test()



# -*- coding: utf-8 -*-
'''
Run me to update the release notes in the documentation.

I will pull the content from ../CHANGE_LOG and insert that into
./source/doc/releases.rst
'''
from __future__ import print_function, division, absolute_import


title_template = '''

.. DO NOT EDIT THIS FILE.
   This file is automatically generated by %(this_script)s.
    

======================
Release Notes
======================

'''

def main():
    from os.path import dirname, join, isfile
    curdir = dirname(__file__)
    changelog_path = join(curdir, '..', 'CHANGE_LOG')
    assert isfile(changelog_path), ("%s does not exist" % changelog_path)
    doc_releases = join(curdir, 'source', 'doc', 'releases.rst')
    with open(changelog_path) as fin:
        content = fin.read()
    with open(doc_releases, 'w') as fout:
        fout.write(title_template % dict(this_script=__file__))
        fout.write(content)

if __name__ == '__main__':
    main()




import time
import itertools

from nose.tools import eq_, ok_
import numpy as np
from .. import extensibletype, methodtable


def test_binsort():
    nbins = 64

    p = np.zeros(nbins, dtype=np.uint16)
    binsizes = np.random.randint(0, 7, size=nbins).astype(np.uint8)

    num_by_size = np.zeros(8, dtype=np.uint16)
    x = np.bincount(binsizes).astype(np.uint16)

    num_by_size[:x.shape[0]] = x
    extensibletype.bucket_argsort(p, binsizes, num_by_size)
    assert np.all(sorted(binsizes) == binsizes[p][::-1])

def test_basic():
    n=64
    prehashes = extensibletype.draw_hashes(np.random, n)
    assert len(prehashes) == len(set(prehashes))
    p, r, m_f, m_g, d = extensibletype.perfect_hash(prehashes, repeat=10**5)
    hashes = ((prehashes >> r) & m_f) ^ d[prehashes & m_g]
    print(p)
    print(d)
    hashes.sort()
    print(hashes)
    assert len(hashes) == len(np.unique(hashes))

# ---
# Test methodtable

def make_signature(type_permutation):
    return "".join(type_permutation[:-1]) + '->' + type_permutation[-1]

def make_ids():
    types = ['f', 'd', 'i', 'l', 'O']
    power = 5
    return map(make_signature, itertools.product(*(types,) * power))

def build_and_verify_methodtable(ids, flags, funcs):
    table = methodtable.PerfectHashMethodTable(methodtable.Hasher())
    table.generate_table(len(ids), ids, flags, funcs)

    for (signature, flag, func) in zip(ids, flags, funcs):
        result = table.find_method(signature)
        assert result is not None

        got_func, got_flag = result
        assert func == got_func, (func, got_func)
        # assert flag == got_flag, (flag, got_flag)

def test_methodtable():
    # ids = ["ff->f", "dd->d", "ii->i", "ll->l", "OO->O"]

    ids = make_ids()
    flags = range(1, len(ids) + 1)
    funcs = range(len(ids))

    step = 100

    i = len(ids)
    for i in range(1, len(ids), step):
        t = time.time()
        build_and_verify_methodtable(ids[:i], flags[:i], funcs[:i])
        t = time.time() - t
        print i, "table building took", t, "seconds."

if __name__ == '__main__':
    test_methodtable()


from extensibletype cimport *

cdef extern from "Python.h":
    ctypedef unsigned int Py_uintptr_t

cdef extern from *:
    ctypedef char *string_t "const char *"

cdef extern from "globalinterning.h":
    ctypedef void *intern_table_t

    intern_table_t *intern_create_table(intern_table_t *table) except NULL
    void intern_destroy_table(intern_table_t *table)
    uint64_t intern_key(intern_table_t *table, string_t key) except? 0

    int PyIntern_Initialize() except -1
    uint64_t PyIntern_AddKey(string_t key) except? 0




rk_randomseed = mtrand.rk_randomseed
rk_randomseed.restype = rk_error
rk_randomseed.argtypes = [rk_address]

def randomseed():
    return rk_randomseed(state_p)

rk_random = mtrand.rk_random
rk_random.restype = ct.c_ulong
rk_random.argtypes = [rk_address]

def random():
    return rk_random(state_p)

rk_long = mtrand.rk_long
rk_long.restype = ct.c_long
rk_long.argtypes = [rk_address]

def long():
    return rk_long(state_p)

rk_ulong = mtrand.rk_ulong
rk_ulong.restype = ct.c_ulong
rk_ulong.argtypes = [rk_address]

def ulong():
    return rk_ulong(state_p)

rk_double = mtrand.rk_double
rk_double.restype = ct.c_double
rk_double.argtypes = [rk_address]

def double():
    return rk_double(state_p)

rk_gauss = mtrand.rk_gauss
rk_gauss.restype = ct.c_double
rk_gauss.argtypes = [rk_address]

def gauss():
    return rk_gauss(state_p)

rk_normal = mtrand.rk_normal
rk_normal.restype = ct.c_double
rk_normal.argtypes = [rk_address,ct.c_double,ct.c_double]

def normal(loc,scale):
    return rk_normal(state_p,loc,scale)

rk_standard_exponential = mtrand.rk_standard_exponential
rk_standard_exponential.restype = ct.c_double
rk_standard_exponential.argtypes = [rk_address]

def standard_exponential():
    return rk_standard_exponential(state_p)

rk_exponential = mtrand.rk_exponential
rk_exponential.restype = ct.c_double
rk_exponential.argtypes = [rk_address,ct.c_double]

def exponential(scale):
    return rk_exponential(state_p,scale)

rk_uniform = mtrand.rk_uniform
rk_uniform.restype = ct.c_double
rk_uniform.argtypes = [rk_address,ct.c_double,ct.c_double]

def uniform(loc,scale):
    return rk_uniform(state_p,loc,scale)

rk_standard_gamma = mtrand.rk_standard_gamma
rk_standard_gamma.restype = ct.c_double
rk_standard_gamma.argtypes = [rk_address,ct.c_double]

def standard_gamma(shape):
    return rk_standard_gamma(state_p,shape)

rk_gamma = mtrand.rk_gamma
rk_gamma.restype = ct.c_double
rk_gamma.argtypes = [rk_address,ct.c_double,ct.c_double]

def gamma(shape,scale):
    return rk_gamma(state_p,shape,scale)

rk_beta = mtrand.rk_beta
rk_beta.restype = ct.c_double
rk_beta.argtypes = [rk_address,ct.c_double,ct.c_double]

def beta(a,b):
    return rk_beta(state_p,a,b)

rk_chisquare = mtrand.rk_chisquare
rk_chisquare.restype = ct.c_double
rk_chisquare.argtypes = [rk_address,ct.c_double]

def chisquare(df):
    return rk_chisquare(state_p,df)

rk_noncentral_chisquare = mtrand.rk_noncentral_chisquare
rk_noncentral_chisquare.restype = ct.c_double
rk_noncentral_chisquare.argtypes = [rk_address,ct.c_double,ct.c_double]

def noncentral_chisquare(df,nonc):
    return rk_noncentral_chisquare(state_p,df,nonc)

rk_f = mtrand.rk_f
rk_f.restype = ct.c_double
rk_f.argtypes = [rk_address,ct.c_double,ct.c_double]

def f(dfnum,dfden):
    return rk_f(state_p,dfnum,dfden)

rk_noncentral_f = mtrand.rk_noncentral_f
rk_noncentral_f.restype = ct.c_double
rk_noncentral_f.argtypes = [rk_address,ct.c_double,ct.c_double,ct.c_double]

def noncentral_f(dfnum,dfden,nonc):
    return rk_noncentral_f(state_p,dfnum,dfden,nonc)

rk_binomial = mtrand.rk_binomial
rk_binomial.restype = ct.c_long
rk_binomial.argtypes = [rk_address,ct.c_long,ct.c_double]

def binomial(n,p):
    return rk_binomial(state_p,n,p)

rk_binomial_btpe = mtrand.rk_binomial_btpe
rk_binomial_btpe.restype = ct.c_long
rk_binomial_btpe.argtypes = [rk_address,ct.c_long,ct.c_double]

def binomial_btpe(n,p):
    return rk_binomial_btpe(state_p,n,p)

rk_binomial_inversion = mtrand.rk_binomial_inversion
rk_binomial_inversion.restype = ct.c_long
rk_binomial_inversion.argtypes = [rk_address,ct.c_long,ct.c_double]

def binomial_inversion(n,p):
    return rk_binomial_inversion(state_p,n,p)

rk_negative_binomial = mtrand.rk_negative_binomial
rk_negative_binomial.restype = ct.c_long
rk_negative_binomial.argtypes = [rk_address,ct.c_double,ct.c_double]

def negative_binomial(n,p):
    return rk_negative_binomial(state_p,n,p)

rk_poisson = mtrand.rk_poisson
rk_poisson.restype = ct.c_long
rk_poisson.argtypes = [rk_address,ct.c_double]

def poisson(lam):
    return rk_poisson(state_p,lam)

rk_poisson_mult = mtrand.rk_poisson_mult
rk_poisson_mult.restype = ct.c_long
rk_poisson_mult.argtypes = [rk_address,ct.c_double]

def poisson_mult(lam):
    return rk_poisson_mult(state_p,lam)

rk_poisson_ptrs = mtrand.rk_poisson_ptrs
rk_poisson_ptrs.restype = ct.c_long
rk_poisson_ptrs.argtypes = [rk_address,ct.c_double]

def poisson_ptrs(lam):
    return rk_poisson_ptrs(state_p,lam)

rk_standard_cauchy = mtrand.rk_standard_cauchy
rk_standard_cauchy.restype = ct.c_double
rk_standard_cauchy.argtypes = [rk_address]

def standard_cauchy():
    return rk_standard_cauchy(state_p)

rk_standard_t = mtrand.rk_standard_t
rk_standard_t.restype = ct.c_double
rk_standard_t.argtypes = [rk_address,ct.c_double]

def standard_t(df):
    return rk_standard_t(state_p,df)

rk_vonmises = mtrand.rk_vonmises
rk_vonmises.restype = ct.c_double
rk_vonmises.argtypes = [rk_address,ct.c_double,ct.c_double]

def vonmises(mu,kappa):
    return rk_vonmises(state_p,mu,kappa)

rk_pareto = mtrand.rk_pareto
rk_pareto.restype = ct.c_double
rk_pareto.argtypes = [rk_address,ct.c_double]

def pareto(a):
    return rk_pareto(state_p,a)

rk_weibull = mtrand.rk_weibull
rk_weibull.restype = ct.c_double
rk_weibull.argtypes = [rk_address,ct.c_double]

def weibull(a):
    return rk_weibull(state_p,a)

rk_power = mtrand.rk_power
rk_power.restype = ct.c_double
rk_power.argtypes = [rk_address,ct.c_double]

def power(a):
    return rk_power(state_p,a)

rk_laplace = mtrand.rk_laplace
rk_laplace.restype = ct.c_double
rk_laplace.argtypes = [rk_address,ct.c_double,ct.c_double]

def laplace(loc,scale):
    return rk_laplace(state_p,loc,scale)

rk_gumbel = mtrand.rk_gumbel
rk_gumbel.restype = ct.c_double
rk_gumbel.argtypes = [rk_address,ct.c_double,ct.c_double]

def gumbel(loc,scale):
    return rk_gumbel(state_p,loc,scale)

rk_logistic = mtrand.rk_logistic
rk_logistic.restype = ct.c_double
rk_logistic.argtypes = [rk_address,ct.c_double,ct.c_double]

def logistic(loc,scale):
    return rk_logistic(state_p,loc,scale)

rk_lognormal = mtrand.rk_lognormal
rk_lognormal.restype = ct.c_double
rk_lognormal.argtypes = [rk_address,ct.c_double,ct.c_double]

def lognormal(mean,sigma):
    return rk_lognormal(state_p,mean,sigma)

rk_rayleigh = mtrand.rk_rayleigh
rk_rayleigh.restype = ct.c_double
rk_rayleigh.argtypes = [rk_address,ct.c_double]

def rayleigh(mode):
    return rk_rayleigh(state_p,mode)

rk_wald = mtrand.rk_wald
rk_wald.restype = ct.c_double
rk_wald.argtypes = [rk_address,ct.c_double,ct.c_double]

def wald(mean,scale):
    return rk_wald(state_p,mean,scale)

rk_zipf = mtrand.rk_zipf
rk_zipf.restype = ct.c_long
rk_zipf.argtypes = [rk_address,ct.c_double]

def zipf(a):
    return rk_zipf(state_p,a)

rk_geometric = mtrand.rk_geometric
rk_geometric.restype = ct.c_long
rk_geometric.argtypes = [rk_address,ct.c_double]

def geometric(p):
    return rk_geometric(state_p,p)

rk_geometric_search = mtrand.rk_geometric_search
rk_geometric_search.restype = ct.c_long
rk_geometric_search.argtypes = [rk_address,ct.c_double]

def geometric_search(p):
    return rk_geometric_search(state_p,p)

rk_geometric_inversion = mtrand.rk_geometric_inversion
rk_geometric_inversion.restype = ct.c_long
rk_geometric_inversion.argtypes = [rk_address,ct.c_double]

def geometric_inversion(p):
    return rk_geometric_inversion(state_p,p)

rk_hypergeometric = mtrand.rk_hypergeometric
rk_hypergeometric.restype = ct.c_long
rk_hypergeometric.argtypes = [rk_address,ct.c_long,ct.c_long,ct.c_long]

def hypergeometric(good,bad,sample):
    return rk_hypergeometric(state_p,good,bad,sample)

rk_hypergeometric_hyp = mtrand.rk_hypergeometric_hyp
rk_hypergeometric_hyp.restype = ct.c_long
rk_hypergeometric_hyp.argtypes = [rk_address,ct.c_long,ct.c_long,ct.c_long]

def hypergeometric_hyp(good,bad,sample):
    return rk_hypergeometric_hyp(state_p,good,bad,sample)

rk_hypergeometric_hrua = mtrand.rk_hypergeometric_hrua
rk_hypergeometric_hrua.restype = ct.c_long
rk_hypergeometric_hrua.argtypes = [rk_address,ct.c_long,ct.c_long,ct.c_long]

def hypergeometric_hrua(good,bad,sample):
    return rk_hypergeometric_hrua(state_p,good,bad,sample)

rk_triangular = mtrand.rk_triangular
rk_triangular.restype = ct.c_double
rk_triangular.argtypes = [rk_address,ct.c_double,ct.c_double,ct.c_double]

def triangular(left,mode,right):
    return rk_triangular(state_p,left,mode,right)

rk_logseries = mtrand.rk_logseries
rk_logseries.restype = ct.c_long
rk_logseries.argtypes = [rk_address,ct.c_double]

def logseries(p):
    return rk_logseries(state_p,p)



# -*- coding: utf-8 -*-

"""
Initial AST validation and normalization.
"""

from __future__ import print_function, division, absolute_import

import ast
import copy

from numba import error
from numba import nodes
from numba import visitors
from numba import typesystem


class NormalizeAST(visitors.NumbaTransformer):
    "Normalize AST"

    function_level = 0

    #------------------------------------------------------------------------
    # Validation
    #------------------------------------------------------------------------

    def visit_GeneratorExp(self, node):
        raise error.NumbaError(
                node, "Generator comprehensions are not yet supported")

    def visit_SetComp(self, node):
        raise error.NumbaError(
                node, "Set comprehensions are not yet supported")

    def visit_DictComp(self, node):
        raise error.NumbaError(
                node, "Dict comprehensions are not yet supported")

    def visit_Raise(self, node):
        raise error.NumbaError(node, "Raise statement not implemented yet")

    #------------------------------------------------------------------------
    # Normalization
    #------------------------------------------------------------------------

    def visit_FunctionDef(self, node):
        if self.function_level:
            return self.handle_inner_function(node)

        self.function_level += 1
        self.visitchildren(node)
        self.function_level -= 1
        return node

    def handle_inner_function(self, node):
        "Create assignment code for inner functions and mark the assignment"
        lhs = ast.Name(node.name, ast.Store())
        ast.copy_location(lhs, node)

        rhs = FuncDefExprNode(func_def=node)
        ast.copy_location(rhs, node)

        fields = rhs._fields
        rhs._fields = []
        assmnt = ast.Assign(targets=[lhs], value=rhs)
        result = self.visit(assmnt)
        rhs._fields = fields

        return result

    def visit_FunctionDef(self, node):
        #for arg in node.args:
        #    if arg.default:
        #        self.visitchildren(arg)
        if self.function_level:
            return self.handle_inner_function(node)

        self.visitchildren(node)
        return node

    def visit_ListComp(self, node):
        """
        Rewrite list comprehensions to the equivalent for loops.

        AST syntax:

            ListComp(expr elt, comprehension* generators)
            comprehension = (expr target, expr iter, expr* ifs)

            'ifs' represent a chain of ANDs
        """
        assert len(node.generators) > 0

        # Create innermost body, i.e. list.append(expr)
        # TODO: size hint for PyList_New
        list_create = ast.List(elts=[], ctx=ast.Load())
        list_create.type = typesystem.object_ # typesystem.ListType()
        list_create = nodes.CloneableNode(list_create)
        list_value = nodes.CloneNode(list_create)
        list_append = ast.Attribute(list_value, "append", ast.Load())
        append_call = ast.Call(func=list_append, args=[node.elt],
                               keywords=[], starargs=None, kwargs=None)

        # Build up the loops from inwards to outwards
        body = append_call
        for comprehension in reversed(node.generators):
            # Hanlde the 'if' clause
            ifs = comprehension.ifs
            if len(ifs) > 1:
                make_boolop = lambda op1_op2: ast.BoolOp(op=ast.And(),
                                                         values=op1_op2)
                if_test = reduce(make_boolop, ifs)
            elif len(ifs) == 1:
                if_test, = ifs
            else:
                if_test = None

            if if_test is not None:
                body = ast.If(test=if_test, body=[body], orelse=[])

            # Wrap list.append() call or inner loops
            body = ast.For(target=comprehension.target,
                           iter=comprehension.iter, body=[body], orelse=[])

        expr = nodes.ExpressionNode(stmts=[list_create, body], expr=list_value)
        return self.visit(expr)

    def visit_AugAssign(self, node):
        """
        Inplace assignment.

        Resolve a += b to a = a + b. Set 'inplace_op' attribute of the
        Assign node so later stages may recognize inplace assignment.

        Do this now, so that we can correctly mark the RHS reference.
        """
        target = node.target

        rhs_target = copy.deepcopy(target)
        rhs_target.ctx = ast.Load()
        ast.fix_missing_locations(rhs_target)

        bin_op = ast.BinOp(rhs_target, node.op, node.value)
        assignment = ast.Assign([target], bin_op)
        assignment.inplace_op = node.op
        return self.visit(assignment)


#------------------------------------------------------------------------
# Nodes
#------------------------------------------------------------------------

class FuncDefExprNode(nodes.Node):
    """
    Wraps an inner function node until the closure code kicks in.
    """

    _fields = ['func_def']


#! /usr/bin/env python
# ______________________________________________________________________
'''Unit test for issue #30.'''
# ______________________________________________________________________
import numpy

from numba import f8
from numba.decorators import jit, autojit
from numba.testing import test_support

import unittest
import __builtin__

# ______________________________________________________________________

def avg2d(arr, result):
    M, N = arr.shape
    for i in range(M):
        avg = 0.
        for j in range(N):
            avg += arr[i,j]
        result[i] = avg / N

# ______________________________________________________________________

def avg2d_w_cast(arr, result):
    M, N = arr.shape
    for i in range(M):
        avg = 0.
        for j in range(N):
            avg += arr[i,j]
        result[i] = avg / float(N)

# ______________________________________________________________________

class TestAvg2DAST(test_support.ASTTestCase):
    def _do_test(self, _avg2d, compiled_fn):
        test_data = numpy.random.random((5,5))
        control_result = numpy.zeros((5,))
        test_result = control_result[:]
        _avg2d(test_data, control_result)
        compiled_fn(test_data, test_result)
        self.assertTrue((control_result == test_result).all())

    def test_avg2d(self):
        compiled_fn = self.jit(argtypes = [f8[:,:], f8[:]])(avg2d)
        self._do_test(avg2d, compiled_fn)

    def test_avg2d_w_cast(self):
        compiled_fn = self.jit(argtypes = [f8[:,:], f8[:]])(avg2d_w_cast)
        self._do_test(avg2d_w_cast, compiled_fn)

# ______________________________________________________________________

if __name__ == "__main__":
#    TestAvg2DAST('test_avg2d_w_cast').debug()
    unittest.main()

# ______________________________________________________________________
# End of test_avg2d.py



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
import llvm.core

from numba import *

class ComplexSupportMixin(object):
    "Support for complex numbers"

    def _generate_complex_op(self, op, arg1, arg2):
        (r1, i1), (r2, i2) = self._extract(arg1), self._extract(arg2)
        real, imag = op(r1, i1, r2, i2)
        return self._create_complex(real, imag)

    def _extract(self, value):
        "Extract the real and imaginary parts of the complex value"
        return (self.builder.extract_value(value, 0),
                self.builder.extract_value(value, 1))

    def _create_complex(self, real, imag):
        assert real.type == imag.type, (str(real.type), str(imag.type))
        complex = llvm.core.Constant.undef(llvm.core.Type.struct([real.type,
                                                                  real.type]))
        complex = self.builder.insert_value(complex, real, 0)
        complex = self.builder.insert_value(complex, imag, 1)
        return complex

    def _promote_complex(self, src_type, dst_type, value):
        "Promote a complex value to value with a larger or smaller complex type"
        real, imag = self._extract(value)

        if dst_type.is_complex:
            dst_type = dst_type.base_type
        dst_ltype = dst_type.to_llvm(self.context)

        real = self.caster.cast(real, dst_ltype)
        imag = self.caster.cast(imag, dst_ltype)
        return self._create_complex(real, imag)

    def _complex_add(self, arg1r, arg1i, arg2r, arg2i):
        return (self.builder.fadd(arg1r, arg2r),
                self.builder.fadd(arg1i, arg2i))

    def _complex_sub(self, arg1r, arg1i, arg2r, arg2i):
        return (self.builder.fsub(arg1r, arg2r),
                self.builder.fsub(arg1i, arg2i))

    def _complex_mul(self, arg1r, arg1i, arg2r, arg2i):
        return (self.builder.fsub(self.builder.fmul(arg1r, arg2r),
                                  self.builder.fmul(arg1i, arg2i)),
                self.builder.fadd(self.builder.fmul(arg1i, arg2r),
                                  self.builder.fmul(arg1r, arg2i)))

    def _complex_div(self, arg1r, arg1i, arg2r, arg2i):
        divisor = self.builder.fadd(self.builder.fmul(arg2r, arg2r),
                                    self.builder.fmul(arg2i, arg2i))
        return (self.builder.fdiv(
            self.builder.fadd(self.builder.fmul(arg1r, arg2r),
                              self.builder.fmul(arg1i, arg2i)),
            divisor),
                self.builder.fdiv(
                    self.builder.fsub(self.builder.fmul(arg1i, arg2r),
                                      self.builder.fmul(arg1r, arg2i)),
                    divisor))

    def _complex_floordiv(self, arg1r, arg1i, arg2r, arg2i):
        real, imag = self._complex_div(arg1r, arg1i, arg2r, arg2i)
        long_type = long_.to_llvm(self.context)
        real = self.caster.cast(real, long_type, unsigned=False)
        imag = self.caster.cast(imag, long_type, unsigned=False)
        real = self.caster.cast(real, arg1r.type, unsigned=False)
        imag = self.caster.cast(imag, arg1r.type, unsigned=False)
        return real, imag




import numba as nb

def py_modulo(restype, argtypes):
    if restype.is_float:
        def py_modulo(a, n):
            r = rem(a, n)

            if (r != 0) and (r < 0) ^ (n < 0):
                r += n

            return r

        instr = 'frem'
    else:
        assert restype.is_int

        def py_modulo(a, n):
            r = rem(a, n)

            if r != 0 and (r ^ n) < 0:
                r += n

            return r

        if restype.is_unsigned:
            instr = 'urem'
        else:
            instr = 'srem'

    rem = nb.declare_instruction(restype(restype, restype), instr)
    return nb.jit(restype(*argtypes))(py_modulo)



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

import ast
import string

import numba
from numba import *
from numba import nodes
from numba import pipeline
from numba import environment
from numba import numbawrapper
from numba.type_inference.module_type_inference import register_value

#------------------------------------------------------------------------
# Intrinsic Classes
#------------------------------------------------------------------------

class Intrinsic(object):

    def __init__(self, func_signature, name):
        self.func_signature = func_signature
        self.name = name

        # Register a type inference function for our intrinsic
        register_infer_intrinsic(self)

        # Build a function wrapper
        self.jitted_func = make_intrinsic(self)

    def __call__(self, *args):
        return self.jitted_func(*args)

    def emit_code(self, lfunc, builder, llvm_args):
        raise NotImplementedError

    def __eq__(self, other):
        return (isinstance(other, type(self)) and
                self.name == other.name and
                self.func_signature == other.func_signature)

    def __hash__(self):
        return hash((type(self), self.name, self.func_signature))


class NumbaInstruction(Intrinsic):

    def emit_code(self, lfunc, builder, llvm_args):
        return getattr(builder, self.name)(*llvm_args)


class NumbaIntrinsic(Intrinsic):

    def emit_code(self, lfunc, builder, llvm_args):
        raise NotImplementedError


def is_numba_intrinsic(value):
    return isinstance(value, Intrinsic)

numbawrapper.add_hash_by_value_type(Intrinsic)

#------------------------------------------------------------------------
# Build Intrinsic Wrapper
#------------------------------------------------------------------------

cache = {}
env = environment.NumbaEnvironment.get_environment()

def make_intrinsic(intrinsic):
    """
    Create an intrinsic function given an Intrinsic.
    """
    if intrinsic in cache:
        return cache[intrinsic]

    # NOTE: don't use numba.jit() and 'exec', it will make inspect.getsource()
    # NOTE: fail, and hence won't work in python 2.6 (since meta doesn't work
    # NOTE: there)

    # Build argument names
    nargs = len(intrinsic.func_signature.args)
    assert nargs < len(string.ascii_letters)
    argnames = ", ".join(string.ascii_letters[:nargs])

    # Build source code and environment
    args = (intrinsic.name, argnames, argnames)
    source = ("def %s(%s): return intrinsic(%s)\n" % args)
    func_globals = {'intrinsic': intrinsic}

    mod_ast = ast.parse(source)
    func_ast = mod_ast.body[0]

    # Compile
    func_env, _ = pipeline.run_pipeline2(
        env, func=None, func_ast=func_ast,
        func_signature=intrinsic.func_signature,
        function_globals=func_globals)
    jitted_func = func_env.numba_wrapper_func

    # Populate cache
    cache[intrinsic] = jitted_func
    return jitted_func

#------------------------------------------------------------------------
# Intrinsic Value Type Inference
#------------------------------------------------------------------------

class IntrinsicNode(nodes.ExprNode):
    "AST Node representing a reference to an intrinsic"

    _fields = ['args']

    def __init__(self, intrinsic, args):
        self.intrinsic = intrinsic
        self.type = self.intrinsic.func_signature.return_type
        self.args = list(args)


def register_infer_intrinsic(intrinsic):
    def infer(*args):
        return IntrinsicNode(intrinsic, args)

    register_value(intrinsic, infer,
                   pass_in_types=False,
                   can_handle_deferred_types=True)

#------------------------------------------------------------------------
# User Exposed Functionality
#------------------------------------------------------------------------

def declare_intrinsic(func_signature, name):
    """
    Declare an intrinsic, e.g.

    >>> declare_intrinsic(void(), "llvm.debugtrap")
    """
    return NumbaIntrinsic(func_signature, name)

def declare_instruction(func_signature, name):
    """
    Declare an instruction, e.g.

    >>> declare_instruction(int32(int32, int32), "add")

    The llvm.core.Builder instruction with the given name will be used.
    """
    return NumbaInstruction(func_signature, name)



from numba import *
import numpy as np

A = np.arange(10, dtype=np.float32)
X = np.empty(10, dtype=np.float32)

@jit(void(f4[:]))
def read_global(Y):
    Y[1] = A[1]

read_global(X)
assert X[1] == 1

@jit(void(f4[:]))
def write_global(Y):
    A[2] = Y[2]

X[2] = 14
write_global(X)
assert A[2] == 14
